using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.ComponentModel;
using System.Windows.Forms;
using Elements.Diagnostic;

namespace Elements
{
    public class ElementControl : UserControl, IControlElement
    {
        protected bool disposed = false, diagnosticMode = false;
        protected string moduleName = "Anonymous_Module";
        protected Port controlPort, diagnosticPort, interfacePort;
        protected List<IPort> ports = new List<IPort>();
        protected object ressourceHandlingLock = new object();

        private Thread ressourceHandlingThread;
        protected Queue<RessourceHandler> ressourceHandlingQueue = new Queue<RessourceHandler>();
        private AutoResetEvent newRessourceHandlerEvent = new AutoResetEvent(false);

        protected ElementControl() { InitElement(); }
        protected ElementControl(string name) { this.moduleName = name; InitElement(); }

        protected virtual void InitElement()
        {
            controlPort = new Port(this, "ControlPort"); ports.Add(controlPort);
            controlPort.AddRessourceHandler(new OperationRessourceHandler("Dispose", this));
            diagnosticPort = new Port(this, "DiagnosticPort"); ports.Add(diagnosticPort);
            interfacePort = new Port(this, "InterfacePort"); ports.Add(interfacePort);
            ressourceHandlingThread = new Thread(new ThreadStart(RessourceHandlingMethod));
            ressourceHandlingThread.IsBackground = true;
            ressourceHandlingThread.Start();

        }
        
        protected override void Dispose(bool dispose)
        {
            if (!disposed)
            {
                if (dispose)
                {
                    foreach (IPort p in ports) { p.Dispose(); }
                    lock (ressourceHandlingThread) { ressourceHandlingThread.Abort(); }
                    base.Dispose(true);
                    return;
                }
                base.Dispose(false);
            }
        }
        public virtual string ModuleName { get { return moduleName; } }
        public virtual List<IPort> Ports { get { return ports; } }
        public virtual IPort ControlPort { get { return controlPort; } }
        public virtual IPort InterfacePort { get { return interfacePort; } }
        public virtual bool DiagnosticMode { get { return diagnosticMode; } set { diagnosticMode = value; } }
        public virtual IPort DiagnosticPort { get { return diagnosticPort; } }

        public virtual void HandleRessource(RessourceHandler resh, IMessage message)
        {
            if (diagnosticMode) { diagnosticPort.AdvertiseRessource("DiagnosticData", new object[1] { message }); }
            lock (ressourceHandlingLock)
            {
                if (!resh.Blocked) //Maybe we need to modify sender about unavailable ressource
                {
                    if (resh.Enqueue) { Enqueue(resh, message); }
                    else { resh.Invoke(message); }
                }
            }
        }

        protected virtual void Enqueue(RessourceHandler r, IMessage m)
        {
            r.DeferInvoke(m);
            lock (ressourceHandlingQueue) { ressourceHandlingQueue.Enqueue(r); }
            newRessourceHandlerEvent.Set();
        }

        private void RessourceHandlingMethod()
        {
            RessourceHandler ressourceHandler;
            
            while (true)
            {
                //try
                {
                    newRessourceHandlerEvent.WaitOne();
                    while (ressourceHandlingQueue.Count != 0)
                    {
                        lock (ressourceHandlingQueue) { ressourceHandler = ressourceHandlingQueue.Dequeue(); }
                        ressourceHandler.AsyncInvoke();
                        Thread.Sleep(0);
                    }
                }
                //catch (Exception e) { }
            }
        }

        public override string ToString() { return moduleName; }

        protected List<IPort> registeredSettingsPort = new List<IPort>();
        public virtual void RegisterWithSettingsPort(IPort settingsPort) { interfacePort.Connect(settingsPort); }
        public virtual void UnRegisterWithSettingsPort(IPort settingsPort) { interfacePort.Disconnect(settingsPort); }
    }
}
