using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Reflection;
using System.Threading;

namespace Elements
{

    public abstract class RessourceHandler
    {
        protected bool noHandling = false, blockOnInvoke = false, enqueue = false, blocked = false;
        protected string ressourceName;
        protected IMessage lastMessage;
        protected object owner;

        public RessourceHandler(object owner) { this.owner = owner; }

        public virtual string RessourceName { get { return ressourceName; } set { ressourceName = value; } }
        public virtual bool BlockOnInvoke { get { return blockOnInvoke; } set { blockOnInvoke = value; } }
        public virtual bool Blocked { get { return blocked; } set { blocked = value; } } 
        public virtual bool Enqueue { get { return enqueue; } set { enqueue = value; } }
        public virtual IMessage LastMessage { get { return lastMessage; } }
        public virtual bool NoHandling{get{return noHandling;}set{noHandling = value;}}

        protected virtual bool Invoke(IResponse response) 
        {
            lastMessage = response; Call(response.Data); return true; 
        }
        public virtual bool DeferInvoke(IMessage message) { lastMessage = message; return true; }
        public virtual bool AsyncInvoke() { bool res = Invoke(lastMessage); return res; }
        public virtual bool Invoke(IMessage message)
        {
            if (blockOnInvoke) { blocked = true; }
            if (message is IRequest) { Invoke((IRequest)message); }
            else if (message is IResponse) { Invoke((IResponse)message); }
            else { blocked = false; return false; }
            blocked = false;
            return true;
        }
        protected virtual bool Invoke(IRequest request)
        {
            lastMessage = request;
            if (noHandling) { Call(request.Args); }
            if (request.Verb == Verb.GET)
            {
                SendResponse(request, new object[1] { Get(request.Args) }, true);
            }
            else if (request.Verb == Verb.POST)
            {
                SendResponse(request, new object[1] { Post(request.Args) }, true);
            }
            else if (request.Verb == Verb.PUT) { if (blockOnInvoke) { return false; } Put(request.Args); }
            else { return false; }
            return true;
        }
        protected abstract void Call(object[] args);
        protected abstract object Get(object[] args);
        protected abstract void Put(object[] args);
        protected abstract object Post(object[] args);
        protected virtual void SendResponse(IRequest request, object[] data, bool success)
        {
            request.Target.Device.SendResponse(request, data, success);
        }
        public override string ToString() { return ressourceName + "Handler"; }
    }

    public class OperationRessourceHandler : RessourceHandler
    {
        protected MethodInfo ressource;

        public OperationRessourceHandler(string name, object owner) : base(owner)
        {
            ressource = owner.GetType().GetMethod(name); ressourceName = ressource.Name;
        }

        protected override void Call(object[] args) { ressource.Invoke(owner, args); }
        protected override object Get(object[] args) { throw new NotImplementedException(); }
        protected override object Post(object[] args) { return ressource.Invoke(owner,args); }
        protected override void Put(object[] args) { ressource.Invoke(owner, args); }
    }

    public class AttributeRessourceHandler : RessourceHandler
    {
        protected PropertyInfo ressource;

        public AttributeRessourceHandler(PropertyInfo ressource, object owner) : base(owner)
        {
            this.ressource = ressource; ressourceName = ressource.Name;
        }
        public AttributeRessourceHandler(string name, object owner): base( owner)
        {
            ressource = owner.GetType().GetProperty(name);
            ressourceName = ressource.Name;
        }

        protected override bool Invoke(IResponse response) { Put(response.Data); return true; }

        protected override void Call(object[] args)
        {
            if (lastMessage is IResponse) { ressource.SetValue(owner, args[0], null); }
            if (lastMessage is IRequest)
            {
                Verb verb = ((IRequest)lastMessage).Verb;
                switch(verb)
                {
                    case Verb.GET: ressource.GetValue(owner, null); break;
                    case Verb.POST: ressource.SetValue(owner, args[0], null); break;
                    case Verb.PUT: ressource.SetValue(owner, args[0], null); break;
                }
            }
        }
        protected override object Post(object[] args) 
        {
            if (args == null) { ressource.SetValue(owner, null, null); return null; }
            ressource.SetValue(owner, args[0], null); return null; 
        }
        protected override void Put(object[] args) 
        {
            if (args == null) { ressource.SetValue(owner, null, null); return; }
            ressource.SetValue(owner, args[0], null); 
        }
        protected override object Get(object[] args) { return ressource.GetValue(owner, null); }
    }
}