using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading;

namespace DataSetTypes
{
    public class DataSet
    {
        protected bool disposed = false;
        protected ushort blockSize;
        protected MemoryStream data;
        protected object objectLock = new object();
        protected string mainUnit = "No Units", secondUnit = "No Units";
        private IParameter<float> masterParameter, slaveParameter;

        public DataSet(MemoryStream data, ushort blockSize) { this.data = data; this.blockSize = blockSize; }
        public DataSet(byte[] buffer, ushort blockSize) { data = new MemoryStream(buffer); this.blockSize = blockSize; }
        public DataSet(MemoryStream data, string mainUnit, string secondUnit, ushort blockSize)
        { 
            this.data = data;
            this.mainUnit = mainUnit;
            this.secondUnit = secondUnit; 
            this.blockSize = blockSize;
        }
        public DataSet(byte[] buffer, string mainUnit, string secondUnit, ushort blockSize)
        { 
            data = new MemoryStream(buffer);
            this.mainUnit = mainUnit;
            this.secondUnit = secondUnit;
            this.blockSize = blockSize;
        }

        public DataSet(byte[] buffer, IParameter<float> master, IParameter<float> slave, ushort blockSize)
        {
            data = new MemoryStream(buffer);
            this.masterParameter = master;
            this.slaveParameter = slave;
            this.blockSize = blockSize;
        }

        public void Dispose() { Dispose(true); }
        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                if (disposing) { data.Dispose(); GC.SuppressFinalize(this); }
                disposed = true;
            }
        }

        public virtual byte[] Buffer
        {
            get { return data.ToArray(); } // Possible optimization there
            set { lock (objectLock) { data = new MemoryStream(value); } }
        }
        public virtual byte this[int index]
        {
            get { data.Position = index; return (byte)data.ReadByte(); }
            set { data.Position = index; data.WriteByte(value); }
        }
        public virtual int Position { get { return (int)data.Position; } set { data.Position = value; } }
        public virtual byte CurrentByte { get { return (byte)data.ReadByte(); } set { data.WriteByte(value); } }
        public virtual int BufferLength { get { return (int)data.Length; } }
        public virtual int BlocksLength { get { return (int)(data.Length / blockSize); } }
        public virtual ushort BlockSize { get { return blockSize; } set { lock (objectLock) { blockSize = value; } } }
        public virtual string MainUnit { get { return mainUnit; } set { mainUnit = value; } }
        public virtual string SecondUnit { get { return secondUnit; } set { secondUnit = value; } }
        public virtual bool Disposed { get { return disposed; } }
        public IParameter<float> MasterParameter { get { return masterParameter; } set { masterParameter = value; } }
        public IParameter<float> SlaveParameter { get { return slaveParameter; } set { slaveParameter = value; } }
        public void Lock() { Monitor.Enter(this); }
        public bool Lock(int timeoutMs) { return Monitor.TryEnter(this, timeoutMs); }
        public void Unlock() { Monitor.Exit(this); }


        public virtual byte[] ReadBytes(int position, int count)
        {
            data.Position = position;
            byte[] read = new byte[count]; data.Read(read, 0, count); return read;
        }
        public virtual object[] ReadBlocks(int blockPosition, int count)
        {
            object[] blocks = new object[count];
            byte[] dataBlock = new byte[blockSize*count];
            
            data.Read(dataBlock, blockPosition / blockSize, dataBlock.Length);
            for (int i = 0; i < count; i++)
            {
                switch (blockSize)
                {
                    case 1: blocks[i] = dataBlock[i]; break;
                    case 2: blocks[i] = BitConverter.ToUInt16(dataBlock, i * 2); break;
                    case 4: blocks[i] = BitConverter.ToUInt32(dataBlock, i * 4); break;
                    case 8: blocks[i] = BitConverter.ToUInt64(dataBlock, i * 8); break;
                }
            }
            return blocks;
        }
    }
}
