using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Motr3D;
using Motr3D.Primitives;
using DataSetTypes;
using Math32.LinearAlgebra;

namespace DataSetVisualizer
{
    public class Axis: LineList
    {
        public static Color DEFAULT_COLOR = Color.Black;
        
        public const float MAX_DISPLACEMENT = Visualizer.NOMINAL_LENGTH / 2;
        public const float MAX_SCALING = 100.0f;
        public const float MIN_SCALING = 0.01f;
        public const int DEFAULT_MAX_DEPTH_LENGTH = 20;
        public const int DEFAULT_MAX_HISTORY_LENGTH = 100;
        
        protected Vertex start, end, direction, normal;
        private Text2D label;
        protected Object3D dataSet;

        protected List<float[]> valuesHistory = new List<float[]>();
        protected float displacement = 0.0f;
        protected float scaling = 1.0f;
        protected volatile int maxDepthLength = DEFAULT_MAX_DEPTH_LENGTH;
        protected volatile int maxHistoryLength = DEFAULT_MAX_HISTORY_LENGTH;
        private string units ="";

        public Axis(Vertex start, Vertex end, Vertex normal, Object3D dataSet, string axisName)
            : base(new float[] { start.x, end.x }, new float[] { start.y, end.y }, new float[] { start.z, end.z })
        {
            this.Color = DEFAULT_COLOR;
            this.end = end;
            this.start = start;
            direction = Vertex.Normalize(end - start);
            this.dataSet = dataSet;
            this.label = new Text2D(axisName+"()", "Arial", 20.0f);
            this.label.Translate(end);
            AddChild(this.label);
            label.Color = Color.Black;
            this.Name = axisName;
            this.normal = Vertex.Normalize(normal);
        }

        public Vertex Start { get { return start; } }
        public Vertex End { get { return end; } }
        public float Length { get { return Vertex.Magnitude(end - start); } }
        public virtual int HistoryCount { get { return valuesHistory.Count; } }
        public string Units 
        { 
            get { return units; } 
            set { units = value; label.CurrentText = Name + "(" + units + ")"; } 
        }
        public override string Name
        {
            get { return base.Name; }
            set 
            {
                base.Name = value;
                label.CurrentText = Name + "(" + units + ")"; 
            }
        }
        public virtual float Displacement 
        { 
            get { return displacement; } 
            set 
            {
                if (Math.Abs(value) >= MAX_DISPLACEMENT) { return; }
                dataSet.Translate(direction * value - direction * displacement);
                displacement = value;
            } 
        }
        public virtual float Scaling
        {
            get { return scaling; }
            set
            {
                Vertex v, vs;                
                if (value > MAX_SCALING || value < MIN_SCALING) { return; }
                v = direction * (value/scaling);
                vs = new Vertex(1, 1, 1);
                vs.x *= 1 + (v.x - direction.x);
                vs.y *= 1 + (v.y - direction.y);
                vs.z *= 1 + (v.z - direction.z);

                dataSet.ScaleVector(vs);

                scaling = value;
            }
        }

        public virtual int Dimensions { get { return 0; } }

        public virtual DataSet CurrentDataSet
        {
            set { }
        }

        public virtual float[][] GetValues(int position)
        {
            int depth = maxDepthLength;

            if (position + depth > valuesHistory.Count) { depth = valuesHistory.Count - position; }
            
            if (position > valuesHistory.Count) { position = valuesHistory.Count - 1; }
            return valuesHistory.GetRange(position, depth).ToArray();            
        }

        public virtual void CreateValues(int length)
        {
            valuesHistory.Insert(0, new float[length]);
             
            if (valuesHistory.Count > maxHistoryLength)
            {
                valuesHistory.RemoveRange(maxHistoryLength, valuesHistory.Count - maxHistoryLength);
            }
            
        }

    }
}
