using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Motr3D;
using Motr3D.Primitives;
using Math32.LinearAlgebra;
using System.Globalization;

namespace DataSetVisualizer
{
    public class AxisGraduation: LineList
    {
        protected const float MAJOR_LENGTH = Visualizer.NOMINAL_LENGTH / 100.0f;
        protected const float MINOR_LENGTH = Visualizer.NOMINAL_LENGTH / 200.0f;
        protected const float FONT_SCALE = 2.0f;

        protected float lowerValue = -1.0f, upperValue = 1.0f;

        protected Vertex normal, lowerBound, upperBound;
        protected float distance, tickLength;
        protected bool enable = true;

        protected List<Text2D> labels = new List<Text2D>();
        protected bool showLabels;
        
        public AxisGraduation(Vertex start, Vertex end, Vertex normal, float distance, float tickLength, bool showLabels)
            :base( new float[]{0,normal.x}, new float[]{0,normal.y}, new float[]{0,normal.z})
        {
            this.normal = Vertex.Normalize(normal);
            this.distance = distance;
            this.tickLength = tickLength;
            this.lowerBound = start;
            this.upperBound = end;
            RefreshGraduations();
            ShowLabels = showLabels;
        }

        public float LowerValue 
        { 
            get { return lowerValue; } 
            set 
            {
                lowerValue = value; 
                RefreshLabels(); 
            }
        }
        public float UpperValue 
        { 
            get { return upperValue; } 
            set 
            { 
                upperValue = value;
                RefreshLabels(); 
            } 
        }

        public virtual float Distance 
        {
            get { return distance; }
            set { distance = value; RefreshLabels(); RefreshGraduations(); } 
        }

        public virtual Vertex LowerBound
        {
            get { return lowerBound; }
            set { lowerBound = value; RefreshGraduations(); RefreshLabels(); }
        }
        public virtual Vertex UpperBound
        {
            get { return upperBound; }
            set { upperBound = value; RefreshGraduations(); RefreshLabels(); }
        }

        public virtual bool ShowLabels
        {
            get { return showLabels; }
            set 
            { 
                showLabels = value;
                if (value) { RefreshLabels(); }
                else { foreach (Text2D t in labels) { t.Dispose(); } labels.Clear(); }
            }
        }
        protected virtual string GenerateLabelString(float val)
        {
            const int MAX_DECIMAL_PLACES = 6;
            
            int decimalPlaces = 1;
            //double decimalPart = val - Math.Round(val, 0);
            val = (float)Math.Round(val, MAX_DECIMAL_PLACES);

            for (int i = MAX_DECIMAL_PLACES; i > 0 ; i--)
            {
                if (val != (float)Math.Round(val, i))
                {
                    val = (float)Math.Round(val, i + 1);
                    decimalPlaces = i + 1;
                    break;
                }
            }
            return val.ToString("F" + decimalPlaces.ToString());
        }
        public virtual void RefreshGraduations()
        {
            int count = (int)(Vertex.Magnitude(upperBound - lowerBound) / distance);
            Vertex translator = lowerBound,
                incrementer = (upperBound - lowerBound) / count;
           

            float[] x = new float[2 * count],
                y = new float[2 * count],
                z = new float[2 * count];
            
            for (int i = 0; i < 2 * count; i += 2)
            {
                x[i] = translator.x; x[i + 1] = x[i] + normal.x * tickLength;
                y[i] = translator.y; y[i + 1] = y[i] + normal.y * tickLength;
                z[i] = translator.z; z[i + 1] = z[i] + normal.z * tickLength;

                translator += incrementer;
            }
            GenerateLineStrip(x, y, z);
            Color = Axis.DEFAULT_COLOR;           
        }

        public virtual void RefreshLabels()
        {
            if (!showLabels) { return; }
            int count = (int)(Vertex.Magnitude(upperBound - lowerBound) / distance);
            float delta = (upperValue - lowerValue) / count;
            Vertex translator = lowerBound
                , incrementer = (upperBound - lowerBound) / count;
            for (int i = 0; i < count; i++)
            {
                float val = lowerValue + delta * i;
                if (labels.Count - 1 < i)
                {
                    labels.Add(new Text2D(GenerateLabelString(val), "Arial", 10.0f));
                    labels[i].Color = Axis.DEFAULT_COLOR;
                    labels[i].Scale(FONT_SCALE);
                    AddChild(labels[i]);
                }
                else { labels[i].CurrentText = GenerateLabelString(val); }
                labels[i].ResetPosition();
                labels[i].Translate( translator);
                translator += incrementer;
            }
            while (labels.Count > count)
            {
                labels[count].Dispose();
                labels.Remove(labels[count]);
            }
        }
    }
}
