using System;
using System.Collections.Generic;
using System.Text;
using DataSetTypes;
using Motr3D;
using Math32.LinearAlgebra;

namespace DataSetVisualizer
{
    public class DataSetAxis : Axis
    {
        protected static DataSet blankDataSet = new DataSet(new byte[100], 1);
        
        private AxisGraduation majorAxisGraduation, minorAxisGraduation;

        public DataSetAxis(Vertex start, Vertex end, Vertex normal, Object3D dataSet, string axisName)
            : base(start, end, normal, dataSet, axisName)
        {
            majorAxisGraduation = new AxisGraduation(start, end, normal, 100, 15, true);
            minorAxisGraduation = new AxisGraduation(start, end, normal, 20, 7.5f, false);
            AddChildren(majorAxisGraduation, minorAxisGraduation);
        }

        public override int Dimensions { get { return 1; } }

        protected virtual AxisGraduation Major { get { return majorAxisGraduation; } }
        protected virtual AxisGraduation Minor { get { return minorAxisGraduation; } }
    }
    
    public class DataSetSlaveAxis: DataSetAxis
    {
        
        protected float upperReference = 1, lowerReference = -1;

        private DataSet currentDataSet = blankDataSet;

        public DataSetSlaveAxis(Vertex start, Vertex end, Vertex normal, Object3D dataSet, string axisName)
            : base(start, end, normal, dataSet, axisName)
        { }

        public override DataSet CurrentDataSet
        {
            set
            {
                
                if (value == null) { return; }
                currentDataSet = value;
                if (Units != value.SlaveParameter.Name) { Units = value.SlaveParameter.Name; }
                if (upperReference != currentDataSet.SlaveParameter.Upper 
                    && lowerReference != currentDataSet.SlaveParameter.Lower)
                {
                    upperReference = currentDataSet.SlaveParameter.Upper;
                    lowerReference = currentDataSet.SlaveParameter.Lower;

                    Major.LowerValue = lowerReference * (1 / scaling);
                    Major.UpperValue = upperReference * (1 / scaling);
                }
            }
        }
        
        

        public override float Scaling
        {
            get { return base.Scaling; }
            set
            {
                base.Scaling = value;
                Major.LowerValue = lowerReference * (1 / scaling);
                Major.UpperValue = upperReference * (1 / scaling);
            }
        }

        public override void CreateValues(int length)
        {
           
            if (currentDataSet == null)
            {
                valuesHistory.Insert(0, new float[length]);
            }
            else { valuesHistory.Insert(0, CreateDataSetValues(currentDataSet, length)); }
                                 
            if (valuesHistory.Count > maxHistoryLength)
            {
                valuesHistory.RemoveRange(maxHistoryLength, valuesHistory.Count - maxHistoryLength);
            }
            
        }

        protected virtual float[] CreateDataSetValues(DataSet dataSet, int length)
        {
            float[] data = new float[length];
            int jump = dataSet.BlocksLength / length;
            float scaler = Length / byte.MaxValue;
            float low = -(Length / 2);
            if (dataSet.Lock(500))
            {
                /*for (int i = 0; i < lenghth; i++)
                {
                    data[i] = ((float)signal[i * signal.BlockSize]
                            + (float)((signal.LowerReference / signal.DeltaReference) * byte.MaxValue))
                            * (float)(Visualizer.NOMINAL_LENGTH / (float)(byte.MaxValue * signal.DeltaReference));
                }*/
                for (int i = 0; i < length; i++)
                {
                    data[i] = low + dataSet[i * jump] * scaler;
                }
                dataSet.Unlock();
            }
            return data;
        }
    }

    public class DataSetMasterAxis : DataSetAxis
    {
        private DataSet currentDataSet = blankDataSet;

        public DataSetMasterAxis(Vertex start, Vertex end, Vertex normal, Object3D dataSet, string axisName)
            : base(start, end, normal, dataSet, axisName)
        { }

        public override DataSet CurrentDataSet
        {
            set
            {
                if (value == null) { return; }
                currentDataSet = value;
                if (Units != currentDataSet.MasterParameter.Name)
                {
                    Units = currentDataSet.MasterParameter.Name;
                }
                if (Major.UpperValue != currentDataSet.MasterParameter.Upper / scaling)
                {      
                    Major.UpperValue = currentDataSet.MasterParameter.Upper / scaling;
                }
                if (Major.LowerValue != currentDataSet.MasterParameter.Lower / scaling)
                {
                    Major.LowerValue = currentDataSet.MasterParameter.Lower / scaling;

                }
            }
        }
        
        public override float Scaling
        {
            get { return base.Scaling; }
            set
            {
                base.Scaling = value;
                Major.UpperValue = currentDataSet.MasterParameter.Upper * value; 
            }
        }
        public override void CreateValues(int length)
        {
            if (currentDataSet == null)
            {
                valuesHistory.Insert(0, new float[length]);
            }
            else { valuesHistory.Insert(0, CreateDataSetValues(currentDataSet, length)); }
                                 
            if (valuesHistory.Count > maxHistoryLength)
            {
                valuesHistory.RemoveRange(maxHistoryLength, valuesHistory.Count - maxHistoryLength);
            }
            
        }

        protected virtual float[] CreateDataSetValues(DataSet dataSet, int length) 
        {
            float[] values = new float[length];
            int increment = dataSet.BlocksLength / length;
            float val = Length / (float)length;
            for (int i = 0; i < length; i++)
            { 
                values[i] = val*i; 
            }

            return values;
        }
    }
}
