using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Threading;
using Motr3D;
using Motr3D.Primitives;
using Elements;
using DataSetTypes;
using Math32.LinearAlgebra;


namespace DataSetVisualizer
{
    public enum DisplayType { LINE, SURFACE };
    public enum DisplayStyle { WIREFRAME, POINTS, SOLID, BARGRAPH };
    public enum DisplayDimensions { ONE, TWO, THREE };
    public enum ColorMode { SINGLE, GRADIENT };
    public enum DisplayModes { NONE, INCREMENT, HISTORY, SAMPLE_MASTER, SAMPLE_SLAVE };

    public abstract class DataSetView: Element
    {
        public const int DEFAULT_WAIT_INTERVAL_MS = 10;
        public const int DEFAULT_MAX_DEPTH = 20;

        private RessourceHandler colorModeResH, singleColorResH, colorGradientResH, displayStyleResH,
            forwardResH, rewindResH, freezeResH, unfreezeResH;
        private List<URL> dataSources = new List<URL>();

        private VertexPrimitive dataSet;
        protected object dataSetLock = new object();
        private Reference axesOrigin = new Reference(), dataSetOrigin = new Reference();
        protected string name = "DataSet", info = "";
        protected DisplayType displayType = DisplayType.LINE;
        protected DisplayStyle displayStyle = DisplayStyle.WIREFRAME;
        private volatile int waitIntervalMs = DEFAULT_WAIT_INTERVAL_MS;
        protected volatile bool freezed = false;
        private volatile int position = 0;
        private byte transparency = byte.MaxValue;
        /*This should be replace with an internal ressource handling message queue*/

        protected int maxHistoryDepth = DEFAULT_MAX_DEPTH;

        private Thread viewUpdateThread;

        private AutoResetEvent updateDisplay = new AutoResetEvent(true);
       
        public event EventHandler Focused;

        protected ColorMode colorMode = ColorMode.SINGLE;
        protected ColorGradient colorGradient = new ColorGradient(
            -Visualizer.NOMINAL_LENGTH / 4,
            Visualizer.NOMINAL_LENGTH / 4,
            Color.Red.ToArgb(),
            Color.Green.ToArgb(),
            Color.Red.ToArgb(), Color.Blue.ToArgb(), Color.Green.ToArgb());
        protected Color singleColor = Color.Blue;

        public DataSetView(string name):base(name)
        { 
            ElementName = name; 
            InitializeAxes();
            InitView();
            SetDisplayStyle(); 
        }

        protected override void InitElement()
        {
            base.InitElement();

            colorModeResH = new AttributeRessourceHandler("ColorMode", controlPort);
            singleColorResH = new AttributeRessourceHandler("SingleColor", controlPort);
            colorGradientResH = new AttributeRessourceHandler("ColorGradient", controlPort);
            displayStyleResH = new AttributeRessourceHandler("DisplayStyle", controlPort);
            forwardResH = new OperationRessourceHandler("Forward", controlPort);
            freezeResH = new OperationRessourceHandler("Freeze", controlPort);
            unfreezeResH = new OperationRessourceHandler("Unfreeze", controlPort);
            rewindResH = new OperationRessourceHandler("Rewind", controlPort);
        }

        protected virtual void InitView()
        {
            viewUpdateThread = new Thread(new ThreadStart(DisplayUpdateLoop));
            viewUpdateThread.IsBackground = false;
            viewUpdateThread.Start();
        }

        protected abstract void InitializeAxes();
 
        //This needs to be properly disposed
        protected override void DisposeElement()
        {
            base.DisposeElement();
            dataSet.Dispose();
            AxesOrigin.Dispose();
            viewUpdateThread.Abort();
        }
        public virtual byte Transparency 
        { 
            get { return transparency; }
            set { transparency = value; if (dataSet != null) { dataSet.Transparency = value; } } 
        }
        public virtual Reference View { get { return dataSetOrigin; } }
        public virtual Reference AxesOrigin { get { return axesOrigin; } }
        public int UpdateWaitIntervalMs { get { return waitIntervalMs; } set { waitIntervalMs = value; } }
        public virtual DisplayType DisplayType
        {
            get { return displayType; }
            set { displayType = value; SetDisplayStyle(); updateDisplay.Set(); }
        }

        public virtual DisplayStyle DisplayStyle
        {
            get { return displayStyle; }
            set
            {
                if (displayStyle == value) { return; }
                displayStyle = value;
                SetDisplayStyle();
                displayStyleResH.Advertise(value); updateDisplay.Set();
            }

        }

        public virtual ColorMode ColorMode
        {
            get { return colorMode; }
            set 
            { 
                colorMode = value;
                colorModeResH.Advertise(value);
                updateDisplay.Set();
            }
        }

        public virtual Color SingleColor 
        { 
            get { return singleColor; } 
            set 
            { 
                singleColor = value; 
                dataSet.Color = value; 
                singleColorResH.Advertise(value);
                updateDisplay.Set();
            } 
        }

        public virtual ColorGradient ColorGradient
        {
            get { return colorGradient; }
            set 
            { 
                colorGradient = value;
                colorGradientResH.Advertise(value);
                updateDisplay.Set();
            }
        }
        public void ForceDisplayUpdate() { updateDisplay.Set(); }
        public virtual void Freeze() { freezed = true; }
        public virtual void Unfreeze() { freezed = false; position = 0; updateDisplay.Set(); }
        public virtual void Forward()
        { 
            if (freezed && position >= 0 && position < maxHistoryDepth ) 
            { 
                position--; 
                updateDisplay.Set();
            }
        }
        public virtual void Rewind() 
        {
            if (freezed && position < maxHistoryDepth && position >= 0) 
            {
                if (position >= maxHistoryDepth) { return; } 
                position++; 
                updateDisplay.Set(); 
            } 
        }

        private void DisplayUpdateLoop()
        {
            float[][][] data3D;
            
            while (true)
            {
                Thread.Sleep(waitIntervalMs);
                updateDisplay.WaitOne();
                data3D = GetDisplayData(position);
                DisplayData(data3D[0], data3D[1], data3D[2]);
            }
        }

        protected abstract float[][][] GetDisplayData(int position);

        public virtual void DisplayData(float[][] xS, float[][] yS, float[][] zS)
        {
            lock (dataSetLock)
            {
                if (dataSet == null) { return; }
                dataSet.Lock();
                if (displayType == DisplayType.LINE)
                {
                    if (dataSet is LineStrip)
                    {
                        LineStrip ls = dataSet as LineStrip;
                        ls.GenerateLineStrip(xS[0], yS[0], zS[0]);
                    }
                    else if (dataSet is Surface)
                    {
                        Surface s = dataSet as Surface;
                        float[] xA;
                        float[] yA;
                        float[] zA;

                        /*if (xS.GetLength(0) < 2) { xA = new float[xS[0].Length]; }
                        else { xA = xS[0]; }
                        if (yS.GetLength(0) < 2) { yA = new float[yS[0].Length]; }
                        else { yA = yS[0]; }
                        if (zS.GetLength(0) < 2) { zA = new float[zS[0].Length]; }
                        else { zA = zS[0]; }*/

                        
                        s.GenerateSurface(
                            new float[2][] { xS[0], new float[xS.GetLength(0)] },
                            new float[2][] { yS[0], new float[yS.GetLength(0)] },
                            new float[2][] { zS[0], new float[zS.GetLength(0)] });
                    }
                    else if (dataSet is PointList)
                    {
                        PointList pl = dataSet as PointList;
                        pl.GeneratePointList(xS[0], yS[0], zS[0]);
                    }
                }
                else if (displayType == DisplayType.SURFACE)
                {
                    if (dataSet is Surface)
                    {
                        if (xS.Length == 0 || yS.Length == 0 || zS.Length == 0) { return; }
                        Surface s = dataSet as Surface;
                        s.GenerateSurface(xS, yS, zS);
                    }
                }
                if (colorMode == ColorMode.GRADIENT)
                {
                    for (int i = 0; i < dataSet.Vertices.Length; i++)
                    {
                        dataSet.Vertices[i].color = colorGradient.GetColorForValue(dataSet.Vertices[i].y);
                    }
                }
                dataSet.Unlock();
            }
        }

        protected virtual void SetDisplayType(int dimensions)
        {
            if (dimensions > 3) { displayType = DisplayType.SURFACE; }
            else { displayType = DisplayType.LINE; }

            SetDisplayStyle();
        }

        protected virtual void SetDisplayStyle()
        {
            if (displayType == DisplayType.LINE)
            {
                if (dataSet != null) { dataSet.Dispose(); }
                switch (displayStyle)
                {
                    case DisplayStyle.BARGRAPH: break;
                    case DisplayStyle.POINTS: dataSet = new PointList(1); break;
                    case DisplayStyle.SOLID:
                        lock (dataSetLock)
                        {
                            dataSet = new Surface(2, 2, 0.01f);
                            ((Surface)dataSet).FillMode = FillMode.SOLID;
                            break;
                        }
                    case DisplayStyle.WIREFRAME: lock (dataSetLock) { dataSet = new LineStrip(1, 2); } break;
                }
            }
            else
            {
                if (!(dataSet is Surface)) { dataSet.Dispose(); dataSet = new Surface(2, 2, 0.01f); }
                switch (displayStyle)
                {
                    case DisplayStyle.BARGRAPH: break;
                    case DisplayStyle.POINTS: ((Surface)dataSet).FillMode = FillMode.POINT; break;
                    case DisplayStyle.SOLID: ((Surface)dataSet).FillMode = FillMode.SOLID; break;
                    case DisplayStyle.WIREFRAME:((Surface)dataSet).FillMode = FillMode.WIREFRAME; break;
                }
            }
            dataSet.Transparency = transparency;
            dataSet.Color = singleColor;
            dataSetOrigin.AddChild(dataSet);
        }

        public void OnFocused() { if (Focused != null) { Focused(this, null); } }

    }
}
