using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Motr3D;
using Motr3D.Primitives;
using Motr3D.Prebuilt;
using Math32.LinearAlgebra;

namespace DataSetVisualizer
{
    public partial class Visualizer : Motr3DViewport
    {
        public const float NOMINAL_LENGTH = 1000.0f;
        protected const float MIN_CAMERA_SCALING_FACTOR = 0.01f;
        protected const float MAX_CAMERA_SCALING_FACTOR = 100f;
        protected const float MAX_CAMERA_DISTANCE = 10 * NOMINAL_LENGTH;
        protected const byte OUT_OF_FOCUS_TRANSPARENCY = 150;

        protected List<DataSetView> dataSetViews = new List<DataSetView>();
        protected DataSetView dataSetViewInFocus;
        protected IMotr3DImplementation motr3DImplementation;

        protected Camera camera;
        protected Reference world = new Reference(), 
            origin = new Reference(), 
            center = new Reference(),
            axisOrigin = new Reference(),
            defaultAxes = new Reference(),
            indicators = new Reference();
        protected Reference screen = new Reference();
        protected float lastMouseX, lastMouseY;
        protected Axis defaultXAxis, defaultYAxis, defaultZAxis;
        protected Sphere[] planIndicators = new Sphere[6] { 
            new Sphere(50,8,8), 
            new Sphere(50,8,8),
            new Sphere(50,8,8),
            new Sphere(50,8,8),
            new Sphere(50,8,8),
            new Sphere(50,8,8)};

        protected Sphere[] lightIndicators = new Sphere[6] { 
            new Sphere(50,8,8), 
            new Sphere(50,8,8),
            new Sphere(50,8,8),
            new Sphere(50,8,8),
            new Sphere(50,8,8),
            new Sphere(50,8,8)};
        protected Light[] lights = new Light[0]
            {

            };

        protected float cameraScaling = 1.0f;
        protected float cameraDistance = NOMINAL_LENGTH;

        public Visualizer(IMotr3DImplementation motr3DImplementation): base(motr3DImplementation)
        {
            InitializeComponent();
            InitDisplay();
        }

        public virtual bool OrthogonalProjection
        {
            get 
            {
                if (world.Projection == Projection.ORTHOGONAL) { return true; }
                return false;
            }
            set 
            {
                if (value) 
                {
                    world.Scale(cameraDistance / NOMINAL_LENGTH);
                    cameraScaling = cameraDistance / NOMINAL_LENGTH;
                    world.Projection = Projection.ORTHOGONAL;
                    return;
                }
                world.Projection = Projection.PERSPECTIVE;
                world.ResetScaling(); cameraScaling = 1.0f;
            }
        }
        public Color BackgroundColor
        {
            get { return camera.BackgroundColor; }
            set 
            { 
                camera.BackgroundColor = value;
                colorDialog.Color = value;
                backgoundColorPanel.BackColor = value;
            }
        }
        public bool ShowAxes 
        { 
            get { return !axisOrigin.BlockRenderLoop; }
            set { axisOrigin.BlockRenderLoop = !value; } 
        }
        public bool ShowIndicators 
        { 
            get { return !indicators.BlockRenderLoop; } 
            set { indicators.BlockRenderLoop = !value; }
        }
        protected override void OnLostFocus(EventArgs e)
        {
            base.OnLostFocus(e);
            if (ContainsFocus) { Focus(); }
        }
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.R) { center.ResetRotation(); center.RotateY(-90); }
            else if (e.KeyCode == Keys.S) { settingsGroupBox.Visible = !settingsGroupBox.Visible; }
            else if (e.KeyCode == Keys.B) { BackgroundColor = Color.Black; }
        }

        /// <summary> 
        /// Trap any keypress before child controls get hold of them
        /// </summary>
        /// <param name="m">Windows message</param>
        /// <returns>True if the keypress is handled</returns>
        protected override bool ProcessKeyPreview(ref Message m)
        {
            const int WM_KEYDOWN = 0x100;
            const int WM_CHAR = 0x102;
            const int WM_SYSCHAR = 0x106;
            const int WM_SYSKEYDOWN = 0x104;
            const int WM_IME_CHAR = 0x286;

            KeyEventArgs e = null;

            if ((m.Msg != WM_CHAR) && (m.Msg != WM_SYSCHAR) && (m.Msg != WM_IME_CHAR))
            {
                e = new KeyEventArgs(((Keys)((int)((long)m.WParam))) | ModifierKeys);
                if ((m.Msg == WM_KEYDOWN) || (m.Msg == WM_SYSKEYDOWN))
                {
                    OnKeyDown(e);
                }
                if (e.Handled) { return e.Handled; }

            }
            return base.ProcessKeyPreview(ref m);
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                center.RotateZ( -((e.Y - lastMouseY) / (this.Height / 2))*90);
                center.RotateY(((e.X - lastMouseX) / (this.Width / 2))*90);
            }
            else if (e.Button == MouseButtons.Right)
            {
                float translate = (e.Y - lastMouseY) * 10;

                if (translate + cameraDistance < MAX_CAMERA_DISTANCE)
                {
                    cameraDistance += translate;
                    cameraDistance = Math.Abs(cameraDistance);
                    camera.Translate(translate, 0, 0);
                }
                if (world.Projection == Projection.ORTHOGONAL)
                {
                    float scale = (e.Y - lastMouseY) / 10 + 1;

                    if (scale * cameraScaling < MAX_CAMERA_SCALING_FACTOR && scale * cameraScaling > MIN_CAMERA_SCALING_FACTOR)
                    {
                        world.Scale(scale);
                        cameraScaling *= scale;
                    }
                }
            }
            
            lastMouseX = e.X; lastMouseY = e.Y;
        }

        protected virtual void DataSetFocused(object sender, EventArgs args)
        {
            DataSetView view = sender as DataSetView;
            
            if (sender == null)  { defaultAxes.BlockRenderLoop = false; return; }
            else { defaultAxes.BlockRenderLoop = true; }
            if (sender.Equals(dataSetViewInFocus)) { return; }
            if (dataSetViewInFocus != null) 
            {
                dataSetViewInFocus.AxesOrigin.BlockRenderLoop = true;
                dataSetViewInFocus.Transparency = OUT_OF_FOCUS_TRANSPARENCY;
            }
            dataSetViewInFocus = sender as DataSetView;
            dataSetViewInFocus.Transparency = byte.MaxValue;
            dataSetViewInFocus.AxesOrigin.BlockRenderLoop = false;
        }

        protected virtual void InitDisplay()
        {
            
            defaultXAxis = new Axis(new Vertex(),
            new Vertex(NOMINAL_LENGTH / 2.0f, 0.0f, 0.0f), new Vertex(), origin, "X");
            defaultYAxis = new Axis(new Vertex(0.0f, -NOMINAL_LENGTH / 2.0f, 0.0f),
            new Vertex(0.0f, NOMINAL_LENGTH / 2.0f, 0.0f), new Vertex(), origin, "Y");
                defaultZAxis = new Axis(new Vertex(),
            new Vertex(0.0f, 0.0f, NOMINAL_LENGTH / 2.0f), new Vertex(), origin, "Z");
            
            camera = new Camera(world);
            camera.BackgroundColor = Color.White;
            this.AddCameras(camera);

            world.Projection = Projection.ORTHOGONAL;
            world.FillMode = FillMode.WIREFRAME;
            world.AddChildren(center);
            world.AddChild(camera);
            
            origin.Translate(-NOMINAL_LENGTH / 4, 0, -NOMINAL_LENGTH / 4 );
            center.AddChildren(origin, indicators);
            defaultAxes.AddChildren(defaultXAxis, defaultYAxis, defaultZAxis);
            defaultXAxis.Visibility = Visibility.CURRENT; 
            defaultYAxis.Visibility = Visibility.CURRENT;
            defaultZAxis.Visibility = Visibility.CURRENT; 
            axisOrigin.AddChild(defaultAxes);
            origin.AddChild(axisOrigin);

            camera.FarClipDistance = 10.0f * NOMINAL_LENGTH; camera.NearClipDistance = 100.0f;
            camera.Translate(-NOMINAL_LENGTH / 1, 0, 0);
            camera.NearClipDistance = 10.0f;
            center.RotateY(-90);

            /*camera.RenderNodes.Add(screen);
            screen.Projection = Projection.ORTHOGONAL;
            screen.AddChild(descriptions);
            screen.Translate(0, Height-20, Width*2.3f);*/

            lightIndicators[0].Color = Color.Yellow; lightIndicators[0].MoveTo(NOMINAL_LENGTH, 0, 0); lightIndicators[0].Transparency = 150;
            lightIndicators[0].Click += new MouseEventHandler(LightIndicatorClicked);
            lightIndicators[1].Color = Color.Yellow; lightIndicators[1].MoveTo(-NOMINAL_LENGTH, 0, 0); lightIndicators[1].Transparency = 150;
            lightIndicators[1].Click += new MouseEventHandler(LightIndicatorClicked);
            lightIndicators[2].Color = Color.Yellow; lightIndicators[2].MoveTo(0, NOMINAL_LENGTH, 0); lightIndicators[2].Transparency = 150;
            lightIndicators[2].Click += new MouseEventHandler(LightIndicatorClicked);
            lightIndicators[3].Color = Color.Yellow; lightIndicators[3].MoveTo(0, -NOMINAL_LENGTH, 0); lightIndicators[3].Transparency = 150;
            lightIndicators[3].Click += new MouseEventHandler(LightIndicatorClicked);
            lightIndicators[4].Color = Color.Yellow; lightIndicators[4].MoveTo(0, 0, NOMINAL_LENGTH); lightIndicators[4].Transparency = 150;
            lightIndicators[4].Click += new MouseEventHandler(LightIndicatorClicked);
            lightIndicators[5].Color = Color.Yellow; lightIndicators[5].MoveTo(0, 0, -NOMINAL_LENGTH); lightIndicators[5].Transparency = 150;
            lightIndicators[5].Click += new MouseEventHandler(LightIndicatorClicked);

            planIndicators[0].Color = Color.Blue; planIndicators[0].MoveTo(NOMINAL_LENGTH - NOMINAL_LENGTH / 10, 0, 0);
            planIndicators[0].Click += new MouseEventHandler(PlanIndicatorClicked);
            planIndicators[1].Color = Color.Blue; planIndicators[1].MoveTo(-(NOMINAL_LENGTH - NOMINAL_LENGTH / 10), 0, 0);
            planIndicators[1].Click += new MouseEventHandler(PlanIndicatorClicked);
            planIndicators[2].Color = Color.Blue; planIndicators[2].MoveTo(0, NOMINAL_LENGTH - NOMINAL_LENGTH / 10, 0);
            planIndicators[2].Click += new MouseEventHandler(PlanIndicatorClicked);
            planIndicators[3].Color = Color.Blue; planIndicators[3].MoveTo(0, -(NOMINAL_LENGTH - NOMINAL_LENGTH / 10), 0);
            planIndicators[3].Click += new MouseEventHandler(PlanIndicatorClicked);
            planIndicators[4].Color = Color.Blue; planIndicators[4].MoveTo(0, 0, NOMINAL_LENGTH - NOMINAL_LENGTH / 10);
            planIndicators[4].Click += new MouseEventHandler(PlanIndicatorClicked);
            planIndicators[5].Color = Color.Blue; planIndicators[5].MoveTo(0, 0, -(NOMINAL_LENGTH - NOMINAL_LENGTH / 10));
            planIndicators[5].Click += new MouseEventHandler(PlanIndicatorClicked);

            indicators.AddChildren(lightIndicators);
            indicators.AddChildren(planIndicators);
            indicators.BlockRenderLoop = true;
        }


        protected virtual void PlanIndicatorClicked(object sender, MouseEventArgs args){}
        protected virtual void LightIndicatorClicked(object sender, MouseEventArgs args){}
        public virtual void AddSignal(DataSetView s) 
        {
           
            dataSetViews.Add(s);
            origin.AddChild(s.View); 
            axisOrigin.AddChild(s.AxesOrigin);
            s.Focused += new EventHandler(DataSetFocused);
            s.Disposed += new EventHandler(DataSetDisposed);
            s.AxesOrigin.BlockRenderLoop = true;
            s.Transparency = OUT_OF_FOCUS_TRANSPARENCY;
        }
        public virtual void RemoveSignal(DataSetView s) 
        {
            dataSetViews.Remove(s);
            origin.RemoveChild(s.View);
            axisOrigin.RemoveChild(s.AxesOrigin);
            s.Transparency = byte.MaxValue;
            s.Focused -= new EventHandler(DataSetFocused);
            s.Disposed -= new EventHandler(DataSetDisposed);
            s.AxesOrigin.BlockRenderLoop = false;
        }

        protected virtual void DataSetDisposed(object sender, EventArgs args)
        {
            RemoveSignal((DataSetView)sender);
        }

        private void backgoundColorButton_Click(object sender, EventArgs e)
        {
            colorDialog.ShowDialog();
            BackgroundColor = colorDialog.Color;
            backgoundColorPanel.BackColor = colorDialog.Color;
        }

        private void orthogonalRadioButton_Click(object sender, EventArgs e)
        {
            OrthogonalProjection = true; perspectiveRadioButton.Checked = false;
        }

        private void perspectiveRadioButton_Click(object sender, EventArgs e)
        {
            OrthogonalProjection = false; orthogonalRadioButton.Checked = false;
        }

        private void indicatorsCheckBox_Click(object sender, EventArgs e)
        {
            ShowIndicators = indicatorsCheckBox.Checked;
        }

        private void axesCheckBox_Click(object sender, EventArgs e)
        {
            ShowAxes = axesCheckBox.Checked;
        }
    }

}
