using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Reflection;
using System.Threading;

namespace Elements
{ 
    public abstract class Element : Component, IElement
    {
        /*This is the default ressource locator for the application domain where this class exists*/
        private static IRessourceLocator domainRessourceLocator;
                
        private bool disposed = false, diagnosticMode = false;
        private string elementName = "Anonymous_Module";
        protected Port diagnosticPort;
        protected List<IPort> ports = new List<IPort>();
        protected object ressourceHandlingLock = new object();
        protected IPort controlPort, ressourceMonitorPort;
        private List<RessourceHandler> ressourceHandlers = new List<RessourceHandler>();
        private RessourceHandler ressourcesLostResH, newRessourcesResH;
        private IRessourceLocator localRessourceLocator;

        private Thread ressourceHandlingThread;
        protected Queue<RessourceHandler> ressourceHandlingQueue = new Queue<RessourceHandler>();
        private AutoResetEvent newRessourceHandlerEvent = new AutoResetEvent(false);

        public event RessourceChangedHandler RessourceChanged;

        //protected Element() { InitElement(); }        
        protected Element(string name) 
        {
            this.elementName = name; InitElement();
            //LocalRessourceLocator = DomainRessourceLocator;
        }
        
        protected virtual void InitElement()
        {
            diagnosticPort = new Port(this, "DiagnosticPort"); ports.Add(diagnosticPort);
            controlPort = new Port(this, "Control"); ports.Add(controlPort);
            ressourceMonitorPort = new Port(this, "RessourceMonitor"); ports.Add(controlPort);
            ressourcesLostResH = new OperationRessourceHandler("RessourcesLost", ressourceMonitorPort);
            newRessourcesResH = new OperationRessourceHandler("NewRessources", ressourceMonitorPort);
            ressourceHandlingThread = new Thread(new ThreadStart(RessourceHandlingMethod));
            ressourceHandlingThread.IsBackground = true;
            ressourceHandlingThread.Start();

        }
        protected override void Dispose(bool managedRessources)
        {
            if (!disposed)
            {
                DisposeElement();
                base.Dispose(managedRessources);
                disposed = true;
            }
        }
        protected virtual void DisposeElement()
        {
            foreach (IPort p in ports) { p.Dispose(); }
            ressourceHandlingThread.Abort();
        }

        public virtual string ElementName { get { return elementName; } protected set { elementName = value; } }
        public virtual List<IPort> Ports { get { return ports; } }
        public virtual IPort ControlPort { get { return controlPort; } }
        public static IRessourceLocator DomainRessourceLocator
        {
            get { return domainRessourceLocator; }
            set { domainRessourceLocator = value; }
        }
        public IRessourceLocator LocalRessourceLocator
        {
            get { return localRessourceLocator; }
            set
            {
                if (localRessourceLocator != null) 
                {
                    ressourceMonitorPort.Quiet(localRessourceLocator.URLRegisterOutput, "LastAddedURLs", "NewRessources");
                    ressourceMonitorPort.Quiet(localRessourceLocator.URLRegisterOutput, "LastRemovedURLs", "RessourcesLost");
                }
                localRessourceLocator = value;
                ressourceMonitorPort.Listen(localRessourceLocator.URLRegisterOutput, "LastAddedURLs", "NewRessources");
                ressourceMonitorPort.Listen(localRessourceLocator.URLRegisterOutput, "LastRemovedURLs", "RessourcesLost");
                NewRessources(localRessourceLocator.URLs);
            }
        }

        public List<RessourceHandler> RessourceHandlers { get { return ressourceHandlers; } }

        public virtual void HandleRessource(RessourceHandler resh, IMessage message) 
        {
            if (diagnosticMode) { diagnosticPort.AdvertiseRessource("DiagnosticData", new object[1] { message }); }
            lock (ressourceHandlingLock)
            {
                if (!resh.Blocked) //Maybe we need to modify sender about unavailable ressource
                {
                    if (resh.Enqueue) { Enqueue(resh, message); }
                    else { resh.Invoke(message); }
                }
            }
        }

        protected virtual void Enqueue(RessourceHandler r, IMessage m) 
        {
            r.DeferInvoke(m);
            if (r == null) { throw new Exception(); }
            ressourceHandlingQueue.Enqueue(r); newRessourceHandlerEvent.Set();
        }

        private void RessourceHandlingMethod()
        {
            RessourceHandler ressourceHandler;

            while (true)
            {
                //try
                {
                    newRessourceHandlerEvent.WaitOne();
                    while (ressourceHandlingQueue.Count > 0)
                    {
                        lock (ressourceHandlingQueue) { ressourceHandler = ressourceHandlingQueue.Dequeue(); }
                        ressourceHandler.AsyncInvoke();
                        Thread.Sleep(0);
                    }
                }
                //catch (Exception e) { }
            }
        }

        public virtual void RessourcesLost(URL[] lostRessources) { }
        public virtual void NewRessources(URL[] newRessources) { }

        public override string ToString() { return elementName; }

        public void OnRessourceChanged(string ressource)
        {
            if (RessourceChanged != null) { RessourceChanged(this, ressource); }
        }
    }
}
