using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Reflection;
using System.Threading;

namespace Elements
{

    public abstract class RessourceHandler
    {
        protected bool blockOnInvoke = false, enqueue = false, blocked = false;
        protected volatile bool concurrencySafe = false;
        protected string ressourceName;
        protected IMessage lastMessage;
        protected IPort[] owners;
        protected IElement element;
        protected URL[] urls;

        public RessourceHandler(string ressourceName, params IPort[] owners) 
        { 
            this.owners = owners; element = owners[0].Owner;
            urls = new URL[owners.Length];
            this.ressourceName = ressourceName;
            for (int i = 0; i < owners.Length; i++) 
            { 
                owners[i].AddRessourceHandlers(this);
                urls[0] = new URL(element.ElementName, owners[i].PortName, ressourceName, owners[i]);
            }
            element.RessourceHandlers.Add(this);
        }

        public string RessourceName { get { return ressourceName; } }
        public virtual bool BlockOnInvoke { get { return blockOnInvoke; } set { blockOnInvoke = value; } }
        public virtual bool Blocked { get { return blocked; } set { blocked = value; } } 
        public virtual bool Enqueue { get { return enqueue; } set { enqueue = value; } }
        public IMessage LastMessage { get { return lastMessage; } }
        public bool ConcurrencySafe { get { return concurrencySafe; } set { concurrencySafe = value; } }
        public URL[] URLs { get { return urls; } }

        protected virtual bool Invoke(IResponse response) 
        {
            lastMessage = response; Put(response.Data); return true; 
        }
        public virtual bool DeferInvoke(IMessage message) { lastMessage = message; return true; }
        public virtual bool AsyncInvoke() { bool res = Invoke(lastMessage); return res; }
        public virtual bool Invoke(IMessage message)
        {
            if (blockOnInvoke) { blocked = true; }
            if (message is IRequest) { Invoke((IRequest)message); }
            else if (message is IResponse) { Invoke((IResponse)message); }
            else { blocked = false; return false; }
            blocked = false;
            return true;
        }
        protected virtual bool Invoke(IRequest request)
        {
            lastMessage = request;
            if (concurrencySafe) { Monitor.Enter(this); }
            switch (request.Verb)
            {
                case Verb.GET: SendResponse(request, new object[1] { Get(request.Args) }, true); break;
                case Verb.POST:
                    SendResponse(request, new object[1] { Post(request.Args) }, true); 
                    element.OnRessourceChanged(ressourceName);
                    break;
                case Verb.PUT: 
                    if (blockOnInvoke) { Monitor.Exit(this); return false; }
                    element.OnRessourceChanged(ressourceName);
                    Put(request.Args); 
                    break;
                default: Monitor.Exit(this); return false; break;
            }            
            Monitor.Exit(this);
            return true;
        }
        protected abstract object Get(object[] args);
        protected abstract void Put(object[] args);
        protected abstract object Post(object[] args);
        public virtual void Advertise(params object[] data)
        {
            for (int i = 0; i < owners.Length; i++) { owners[i].AdvertiseRessource(ressourceName, data); }
            element.OnRessourceChanged(ressourceName);
        }
        protected virtual void SendResponse(IRequest request, object[] data, bool success)
        {
            request.Target.Device.SendResponse(request, data, success);
        }
        public override string ToString() { return ressourceName + " Handler"; }
    }

    public class OperationRessourceHandler : RessourceHandler
    {
        protected MethodInfo ressource;

        public OperationRessourceHandler(string name,params IPort[] owners) : base(name,owners)
        {
            ressource = element.GetType().GetMethod(name);
            if (ressource == null) { throw new ArgumentException("Ressource not found"); }
        }
        public OperationRessourceHandler(string name, object target, params IPort[] owners)
            : base(name, owners)
        {
            ressource = target.GetType().GetMethod(name);
            if (ressource == null) { throw new ArgumentException("Ressource not found"); }
        }

        protected override object Get(object[] args) { throw new NotImplementedException(); }
        protected override object Post(object[] args) { return ressource.Invoke(element,args); }
        protected override void Put(object[] args) { ressource.Invoke(element, args); }
    }

    public class AttributeRessourceHandler : RessourceHandler
    {
        protected PropertyInfo ressource;

        public AttributeRessourceHandler(string name, params IPort[] owners): base(name, owners)
        {
            ressource = element.GetType().GetProperty(name);
            if (ressource == null) { throw new ArgumentException("Ressource not found"); }
        }
        public AttributeRessourceHandler(string name, object target, params IPort[] owners)
            : base(name,owners)
        {
            ressource = target.GetType().GetProperty(name);
            if (ressource == null) { throw new ArgumentException("Ressource not found"); }
        }

        protected override bool Invoke(IResponse response) { Put(response.Data); return true; }

        protected override object Post(object[] args) 
        {
            if (args == null) { ressource.SetValue(element, null, null); return null; }
            ressource.SetValue(element, args[0], null); return null; 
        }
        protected override void Put(object[] args) 
        {
            if (args == null) { ressource.SetValue(element, null, null); return; }
            ressource.SetValue(element, args[0], null); 
        }
        protected override object Get(object[] args) { return ressource.GetValue(element, null); }
    }
}