using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using DataSetTypes;
using Elements;

namespace SignalProcessing
{
    public class Trigger :SignalProcessor
    {
        protected volatile float preTriggerRatio = 0, postTriggerRatio = 0.5f, levelRatio = 0;
        protected volatile bool isRisingEdge = true, isExternalTrigger = false;
        protected volatile bool triggered = false, enable = true;
        protected IPort triggerInputPort, triggerOutputPort;
        protected int lastTriggeredIndex = 0;

        private RessourceHandler preTriggerRatioResH,
            postTriggerRatioResH,
            levelRatioResH,
            isRisingEdgeResH,
            externalTriggerResH,
            enableResH;

        public Trigger(string name) : base( name) { }

        protected override void InitElement()
        {
            base.InitElement();
            triggerInputPort = new Port(this, "TriggerInput"); ports.Add(triggerInputPort);
            triggerOutputPort = new Port(this, "TriggerOutput"); ports.Add(triggerOutputPort);

            preTriggerRatioResH = new AttributeRessourceHandler("PreTriggerRatio", controlPort);
            postTriggerRatioResH = new AttributeRessourceHandler("PostTriggerRatio", controlPort);
            levelRatioResH = new AttributeRessourceHandler("LevelRatio", controlPort);
            isRisingEdgeResH = new AttributeRessourceHandler("IsRisingEdge", controlPort);
            externalTriggerResH = new AttributeRessourceHandler("ExternalTrigger", controlPort);
            enableResH = new AttributeRessourceHandler("Enable", controlPort);
            
        }

        public float PreTriggerRatio
        {
            get { return preTriggerRatio; }
            set 
            { 
                preTriggerRatio = value;
                preTriggerRatioResH.Advertise(value); 
            }
        }
        public float PostTriggerRatio
        {
            get { return postTriggerRatio; }
            set 
            { 
                postTriggerRatio = value;
                postTriggerRatioResH.Advertise(value);
            }
        }
        public float LevelRatio
        {
            get { return levelRatio; }
            set 
            {
                levelRatio = value;
                levelRatioResH.Advertise(value);
            }
        }
        public bool IsRisingEdge
        {
            get { return isRisingEdge; }
            set 
            { 
                isRisingEdge = value;
                isRisingEdgeResH.Advertise(value);
            }
        }
        public bool ExternalTrigger
        {
            get { return isExternalTrigger; }
            set 
            {
                isExternalTrigger = value;
                externalTriggerResH.Advertise(value);
            }
        }
        public bool Triggered { get { return triggered; } }

        public bool Enable { get { return enable; } set { enable = value; } }

        public virtual void FireTrigger() { if (!isExternalTrigger) { return; } triggered = true; }

        public override DataSet ProcessSignal(DataSet signal)
        {
            byte level = (byte)(levelRatio*byte.MaxValue);
            int triggeredIndex = -1;
            bool sync_isRisingEdge = isRisingEdge;
            byte lastValue, currentValue;

            if (!enable) { return signal; }

            if(signal.BlockSize * signal.Buffer.Length < 2)
            {
                throw new ArgumentException("Signal.Buffer.Length < 2");
            }

            if ((lastTriggeredIndex != 0) || (lastTriggeredIndex < (signal.BufferLength / signal.BlockSize)))
            {

            }

            //Will depend if platform is little-endian or not
            lastValue = signal[signal.BlockSize-1];
            currentValue = signal[2*signal.BlockSize-1];

            for (int i = 3; i < signal.BlocksLength; i++) 
            {
                if (triggered) { triggeredIndex = i; break; }
                if ((lastValue < level) && (currentValue>level))
                {
                    if (sync_isRisingEdge) { triggeredIndex = i; break; }
                }
                else if ((lastValue > level) && (currentValue < level)) { if (!sync_isRisingEdge) { triggeredIndex = i; break; } }
                else if ((lastValue == level) || (currentValue == level)) { triggeredIndex = i; break; } //Need More Precision

                lastValue = currentValue;
                currentValue = signal[i * signal.BlockSize - 1];
            }
            if (triggeredIndex == -1) { return null; }

            /*int preTrigger = (int)(preTriggerRatio * triggeredIndex * signal.BlockSize)
                , postTrigger = (int)(postTriggerRatio * (signal.BufferLength - triggeredIndex * signal.BlockSize) );*/

            int preTrigger = (int)(preTriggerRatio * signal.BlocksLength)
                , postTrigger = (int)(postTriggerRatio * signal.BlocksLength);

            if (postTrigger > signal.BlocksLength - triggeredIndex * signal.BlockSize) 
            {
                postTrigger = signal.BlocksLength - triggeredIndex * signal.BlockSize;
            }
            if (preTrigger > triggeredIndex * signal.BlockSize)
            {
                preTrigger = triggeredIndex * signal.BlockSize;
            }

            byte[] buffer = new byte[preTrigger + postTrigger];
            if (signal.Lock(20))
            {
                signal.ReadBytes(triggeredIndex * signal.BlockSize - preTrigger, buffer.Length - 1).CopyTo(buffer, 0);
                signal.Unlock();
            }
            else { return null; }

            //triggerOutputPort.AdvertiseRessource("Trigger");
            triggered = false;

            return new DataSet(
                buffer,
                new Parameter<float>(signal.MasterParameter.Upper * (buffer.Length / signal.BufferLength),
                    signal.MasterParameter.Lower * (buffer.Length / signal.BufferLength),
                    signal.MasterParameter.Name),
                signal.SlaveParameter,
                signal.BlockSize);

        }
    }
}
