using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.IO;
using Driver;
using Elements;
using DataSetTypes;


namespace VirtualSignalGeneratorDriver
{
    
    public class VirtualSignalGenerator: AbstractDriver
    {
        private IPort outputPort;
        private RessourceHandler 
            outputSignalResH, 
            lowRefResH, 
            upRefResH, 
            signalTypeResH,
            samplesPerPeriodResH,
            periodLengthMsResH,
            intervalMsResH;

        private SignalType signalType = SignalType.NONE;
        private ISignalGenerator signal =  new Sine();
        protected Timer timer;
        protected Type sampleType = typeof(byte);
        AutoResetEvent generateEvent = new AutoResetEvent(true);
        private int intervalMs = 100;
        private int samplesPerPeriod = 100;
        private int periodLengthMs = 10;
        private int periodNumber = 1000;
        private float lowerRef = -1f, upperRef = 1f;
        private volatile DataSet outputSignal;
        private Thread signalGeneratorThread;


        public VirtualSignalGenerator(): base("VirtualSignalGenerator")
        {
            timer = new Timer(new TimerCallback(TimerCallBack));
            timer.Change(0, intervalMs);
            signalGeneratorThread = new Thread(new ThreadStart(GenerateSignal));
            signalGeneratorThread.IsBackground = false;
            signalGeneratorThread.Start();
        }

        protected override void InitElement()
        {
            base.InitElement();
            outputPort = new Port(this, "OutputPort");
            outputSignalResH = new AttributeRessourceHandler("OutputSignal", outputPort);
            lowRefResH = new AttributeRessourceHandler("LowerReference", ControlPort);
            upRefResH = new AttributeRessourceHandler("UpperReference", ControlPort);
            signalTypeResH = new AttributeRessourceHandler("SignalType", ControlPort);
        }

        public IPort OutputPort { get { return outputPort; } }

        public DataSet OutputSignal
        {
            get { return outputSignal; }
            set
            {
                outputSignal = value;
                outputSignalResH.Advertise(value);
            }
        }

        public SignalType SignalType
        {
            get { return signalType; }
            set 
            { 
                signalType = value;
                switch (signalType)
                {
                    case SignalType.SINE: signal = new Sine(); break;
                    case SignalType.COSINE: signal = new Cosine(); break;
                    case SignalType.SQUARE: signal = new Square(); break;
                    case SignalType.TRIANGLE: signal = new Triangle(); break;
                }
                generateEvent.Set();
            }
        }

        public int IntervalMs 
        { 
            get { return intervalMs; }
            set { intervalMs = value; timer.Change(0, value); } 
        }
        public int SamplesPerPeriod
        {
            get { return samplesPerPeriod; }
            set { samplesPerPeriod = value; generateEvent.Set(); }
        }
        public int PeriodLengthMs
        {
            get { return periodLengthMs; }
            set { periodLengthMs = value; generateEvent.Set(); }
        }
        public int NumberOfSamples
        {
            get { return periodNumber; }
            set { periodNumber = value; generateEvent.Set(); }
        }
        public float LowerReference
        {
            get { return lowerRef; }
            set 
            { 
                lowerRef = value;
                if (outputSignal != null) { outputSignal.SlaveParameter.Lower = lowerRef; }
                lowRefResH.Advertise(value);
            }
        }
        public float UpperReference
        {
            get { return upperRef; }
            set
            {
                upperRef = value;
                if (outputSignal != null) { outputSignal.SlaveParameter.Upper = upperRef; }
                upRefResH.Advertise(value);
            }
        }
        
        private void GenerateSignal()
        {
            byte[] output, period;

            while (true)
            {
                generateEvent.WaitOne();
                ISignalGenerator signal = this.signal; 
                int intervalMs = this.intervalMs;
                int samplesPerPeriod = this.samplesPerPeriod;
                int periodLengthMs = this.periodLengthMs;
                int periodNumber = this.periodNumber;
                float lowerRef = this.lowerRef, upperRef = this.upperRef;
              
                period = signal.GenerateSignal(samplesPerPeriod);
                output = new byte[periodNumber];
                for (int i = 0, j = 0; i < periodNumber; i++, j++)
                {
                    if (j > samplesPerPeriod - 1) { j = 0; }
                    output[i] = period[j];
                }

                outputSignal = new DataSet(
                    output,
                    new Parameter<float>((float)periodLengthMs * (float)(periodNumber / samplesPerPeriod), 0, "Seconds"),
                    new Parameter<float>(upperRef, lowerRef, "Volts"),
                    1);
                Thread.Sleep(0);
            }
        }
        

        public virtual void TimerCallBack(Object state)
        {
            if (State == States.STOPPED) { return; }
            if (outputSignal != null) { OutputSignal = outputSignal; }
        }
    }
}
