using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.IO;
using DataSetTypes;
using Driver;
using Elements;

namespace SoundCardInputDriver
{
    public class SoundInputChannel: AbstractDriver
    {
        protected SoundInputDriver owner;
        protected Rates currentRate;
        protected Bps bitsPerSample;
        protected int bufferSize;
        protected PcmSample currentPcmSample;
        protected volatile bool invalid = false;
        protected DataSet analogSample;
        protected object sampleLock = new object();
        protected Port samplePort;
        private RessourceHandler currentRateResH, bufferSizeResH, bitsPerSampleResH, sampleResH;

        //channelnumber must be zero based
        protected internal SoundInputChannel(SoundInputDriver owner, uint channelNumber)
            : base(owner.Name + "_Ch_" + (channelNumber + 1))
        {
            interfaceNumber = channelNumber;
            this.owner = owner;
            owner.OnNewPcmSample += new NewPcmSampleHandler(NewPcmSample);
            currentRate = owner.CurrentRate;
            bufferSize = owner.BufferSize;
            bitsPerSample = owner.BitsPerSample;
        }

        protected override void InitElement()
        {
            base.InitElement();
            samplePort = new Port(this, "SamplePort");
            currentRateResH = new AttributeRessourceHandler("CurrentRate", controlPort);
            bufferSizeResH = new AttributeRessourceHandler("BufferSize", controlPort);
            bitsPerSampleResH = new AttributeRessourceHandler("BitsPerSample", controlPort);
            sampleResH = new AttributeRessourceHandler("Sample", samplePort);
            sampleResH.Enqueue = true; sampleResH.BlockOnInvoke = true;
        }

        public virtual Rates CurrentRate
        {
            get { return currentRate; }
            set
            {
                lock (propertiesLock) { currentRate = value; }
                owner.CurrentRate = value;
                currentRateResH.Advertise(value);
            }
        }
        public virtual int BufferSize
        {
            get { return bufferSize; }
            set
            {
                lock (propertiesLock) { bufferSize = value; }
                owner.BufferSize = value;
                bufferSizeResH.Advertise(bufferSize);
            }
        }
        public virtual Bps BitsPerSample
        {
            get { return bitsPerSample; }
            set
            {
                lock (propertiesLock) { bitsPerSample = value; }
                owner.BitsPerSample = value;
                bitsPerSampleResH.Advertise(value);
            }
        }

        public virtual Port SamplePort { get { return samplePort; } }
        public virtual DataSet Sample { get { return analogSample; } }

        protected virtual void NewPcmSample(PcmSample sample)
        {
            lock (objectLock) { if (state == States.STOPPED) { return; } }
            lock (sampleLock) { currentPcmSample = sample; }
            invalid = true;
            analogSample = Convert(currentPcmSample);
            sampleResH.Advertise(analogSample);
        }
        public override void Start() 
        {
            lock (objectLock) { state = States.STARTED; }
            BitsPerSample = bitsPerSample;
            BufferSize = bufferSize;
            CurrentRate = currentRate;  
            owner.Start();
        }
        public override void Stop() 
        { 
            lock (objectLock) { state = States.STOPPED; }
            owner.Stop();
        }

        protected virtual DataSet Convert(PcmSample s)
        {
            int stride = 0, startOffset, bytesPerSample = (int)bitsPerSample / 8, buffSize;

            stride += (int)s.rate / (int)currentRate;
            stride *= (int)bitsPerSample / (int)s.bps;
            stride *= (int)s.channels;
            startOffset = (int)(stride * interfaceNumber);

            if (stride * bufferSize > s.length) { buffSize = s.length / stride; }
            else { buffSize = bufferSize; }
            s.sample.Position = startOffset;

            byte[] buffer = new byte[bufferSize * bytesPerSample], sample = new byte[s.sample.Length]; 
            s.sample.Read(sample, 0, (int)(buffSize* stride - startOffset));

            for (int i = 0, j = 0; i < buffSize; i+=bytesPerSample, j += stride)
            {
                buffer[i] = sample[j];
                if (bitsPerSample > s.bps) { buffer[i + 1] = sample[j + 1]; }
            }

            return new DataSet(
                buffer,
                new Parameter<float>(1 / (float)currentRate * buffSize, 0, "Seconds"),
                new Parameter<float>(0.7f, -0.7f, "Volts"),
                (ushort)bytesPerSample);
        }

    }
}
