using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.IO;
using Driver;
using Microsoft.DirectX.DirectSound;
using DataSetTypes;
namespace SoundCardInputDriver
{
    public enum Rates{R_11KHz = 11025, R_22KHz=22050, R_44KHz=44100, R_48KHz=48000, R_96KHz=96000};
    public enum Bps{B_8Bits = 8, B_16Bits = 16};
    public enum Channels { MONO = 1, STEREO = 2 };

    public delegate void NewPcmSampleHandler(PcmSample currentSample);

    public class SoundInputDriver
    {        
        protected DeviceInformation devInfo;
        protected CaptureCaps captureCaps;
        protected Capture capture;
        protected CaptureBufferDescription captureBufferDescription;
        protected CaptureBuffer captureBuffer;
        protected AutoResetEvent bufferFullEvent = new AutoResetEvent(false);
        protected States state = States.STOPPED;
        protected Notify notify;
        protected Thread waitingThread;
        protected PcmSample currentPcmSample;
        protected object objectLock = new object();
        protected object sampleLock = new object();
        protected object descriptionLock = new object();
        protected bool disposed = false;
        protected List<SoundInputChannel> registeredSoundInputChannels = new List<SoundInputChannel>();

        protected internal event NewPcmSampleHandler OnNewPcmSample;

        private Bps bitsPerSample = Bps.B_8Bits;
        private Rates rate = Rates.R_11KHz;
        private int bufferSize = 128;
        private Channels channels = Channels.MONO;

        protected internal SoundInputDriver(DeviceInformation devInfo)
        {
            this.devInfo = devInfo;
            capture = new Capture(devInfo.DriverGuid);
            captureCaps = capture.Caps;
            if (captureCaps.Channels == (int)Channels.STEREO) { channels = Channels.STEREO; }
            waitingThread = new Thread(new ThreadStart(WaitForBufferFull));
            waitingThread.IsBackground = true;
            waitingThread.Start();
        }
        public virtual string Name { get { return devInfo.Description; } }
        
        public virtual void Dispose() { Dispose(true); }
        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                disposed = true;
                if (disposing) 
                { 
                    this.Stop();
                    captureBuffer.Dispose();
                    capture.Dispose();
                    notify.Dispose();
                }
            }
        }

        
        public void RegisterSoundInputChannel(SoundInputChannel s) { registeredSoundInputChannels.Add(s); }

        public bool Disposed { get { return disposed; } }
        public virtual Rates CurrentRate
        {
            get { return rate; }
            protected internal set { UpdateAcquisitionParameters(); }
        }
        public virtual Bps BitsPerSample
        {
            get { return bitsPerSample; }
            protected internal set { UpdateAcquisitionParameters(); }
        }
        public virtual int BufferSize
        {
            get { return bufferSize; }
            protected internal set { UpdateAcquisitionParameters(); }

        }
        public virtual States State { get { return state; } }
        public virtual Channels Channels
        {
            get { return channels; }
            protected internal set { UpdateAcquisitionParameters(); }
        }
        
        public ushort BlockAlign
        {
            get { return (ushort)captureBufferDescription.Format.BlockAlign; }
        }
        protected internal PcmSample CurrentPcmSample { get { lock (sampleLock) { return currentPcmSample; } } }

        protected virtual void UpdateAcquisitionParameters()
        {
            int bufferSize = 0;
            Bps bps = Bps.B_8Bits;
            Rates rate = Rates.R_11KHz;

            lock (registeredSoundInputChannels)
            {
                foreach (SoundInputChannel s in registeredSoundInputChannels)
                {
                    if (s.BitsPerSample > bps) { bps = s.BitsPerSample; }
                    if (s.CurrentRate > rate) { rate = s.CurrentRate; }
                    if (((int)rate / (int)s.CurrentRate) * s.BufferSize > bufferSize)
                    {
                        bufferSize = ((int)rate / (int)s.CurrentRate) * s.BufferSize;
                    }
                }
            }

            lock (descriptionLock)
            {
                this.bufferSize = bufferSize;
                this.bitsPerSample = bps;
                this.rate = rate;
            }

            StopAcquisition();
            if (state == States.STARTED) { StartAcquisition(); }
        }

        public virtual void Start() 
        {
            if (state == States.STARTED) { return; } 
            state = States.STARTED;
            UpdateAcquisitionParameters(); 
        }
        protected virtual void StartAcquisition()
        {
            CaptureBufferDescription cbd = new CaptureBufferDescription(); 
            WaveFormat format = new WaveFormat();
            
            lock (objectLock)
            {
                cbd.BufferBytes = bufferSize * ((int)bitsPerSample / 8) * (int)channels;
                format.AverageBytesPerSecond = (int)rate * ((int)bitsPerSample / 8) * (int)channels;
                format.BitsPerSample = (short)bitsPerSample;
                format.BlockAlign = (short)(((short)bitsPerSample / 8) * (short)channels);
                format.Channels = (short)channels;
                format.FormatTag = WaveFormatTag.Pcm;
                format.SamplesPerSecond = (int)rate;
                cbd.WaveMapped = false;
                cbd.Format = format;
                try
                {
                    if (notify != null) { notify.Dispose(); }
                    if (captureBuffer != null) { captureBuffer.Dispose(); }
                }
                catch (NullReferenceException) { }

                captureBuffer = new CaptureBuffer(cbd, capture);
                notify = new Notify(captureBuffer);

                BufferPositionNotify[] bnArr = new BufferPositionNotify[1];
                bnArr[0].Offset = cbd.BufferBytes - 1;
                bnArr[0].EventNotifyHandle = bufferFullEvent.Handle;
                notify.SetNotificationPositions(bnArr);

                captureBuffer.Start(true);
                state = States.STARTED;
            }
        }

        public virtual void Stop() 
        {
            foreach (SoundInputChannel s in registeredSoundInputChannels) { if (s.State == States.STARTED) { return; } }
            lock (objectLock) { state = States.STOPPED; }
            StopAcquisition();
        }
        protected virtual void StopAcquisition()
        { 
            lock (objectLock)
            {
                if (captureBuffer != null) { captureBuffer.Stop(); }
            }
        }
        
        private void WaitForBufferFull()
        {
            MemoryStream stream;
            PcmSample s;

            while (true)
            {
                bufferFullEvent.WaitOne();
                lock (descriptionLock)
                {
                    s.length = bufferSize * ((int)bitsPerSample / 8) * (int)channels;
                    s.rate = CurrentRate;
                    s.bps = BitsPerSample;
                    s.blockAlign = (ushort)(((short)bitsPerSample / 8) * (short)channels);
                    s.channels = Channels;
                    
                    stream = new MemoryStream();
                    try
                    {
                        captureBuffer.Read(
                            0,
                            stream,
                           bufferSize * ((int)bitsPerSample / 8) * (int)channels,
                            LockFlag.None);
                    }
                    catch (Exception) { }
                }
                s.sample = stream;
                lock(sampleLock) { currentPcmSample = s;  }
                OnNewPcmSample(s); // Synchronous Processing...
                stream.Dispose();

            }
        }
    }
}
