using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using DataSetVisualizer;
using Elements;
using DataSetTypes;
using Motr3D;
using Motr3D.Primitives;
using Math32.LinearAlgebra;
using System.Threading;

namespace XYZDataSetVisualizer
{
    public enum AxisName { X = 0, Y = 1, Z = 2 };
    
    public class XYZDataSetView: DataSetView
    {
        private IPort xPort, yPort, zPort;
        private RessourceHandler xResH, yResH, zResH,
            xDisplacementResH, yDisplacementResH, zDisplacementResH,
            xScalingResH, yScalingResH, zScalingResH
            , dataSourcesResH, xDataSourceResH, yDataSourceResH, zDataSourceResH;
        private object displayUpdateLock = new object();
        private volatile DataSet currentX, currentY, currentZ;
        protected Axis xAxis, yAxis, zAxis;
        protected DisplayModes xAxisDisplayMode = DisplayModes.NONE,
           yAxisDisplayMode = DisplayModes.NONE,
           zAxisDisplayMode = DisplayModes.NONE;
        protected List<URL> dataSources = new List<URL>();
        protected URL xDataSource, yDataSource, zDataSource;

        protected Thread viewUpdateThread;

        public XYZDataSetView(string name) : base(name) 
        {
            //DataSources.AddRange(RessourceLocator.SearchRessource("ChannelOutput"));
        }

        protected override void InitElement()
        {
            base.InitElement();
            xPort = new Port(this, "XPort");
            yPort = new Port(this, "YPort");
            zPort = new Port(this, "ZPort");

            xResH = new AttributeRessourceHandler("X", xPort);
            xResH.Enqueue = true;
            yResH = new AttributeRessourceHandler("Y", yPort);
            yResH.Enqueue = true;
            zResH = new AttributeRessourceHandler("Z", zPort);
            zResH.Enqueue = true;
            xDisplacementResH = new AttributeRessourceHandler("XDisplacement", ControlPort);
            yDisplacementResH = new AttributeRessourceHandler("YDisplacement", ControlPort);
            zDisplacementResH = new AttributeRessourceHandler("ZDisplacement", ControlPort);

            xScalingResH = new AttributeRessourceHandler("XScaling", ControlPort);
            yScalingResH = new AttributeRessourceHandler("YScaling", ControlPort);
            zScalingResH = new AttributeRessourceHandler("ZScaling", ControlPort);
            dataSourcesResH = new AttributeRessourceHandler("DataSources", ControlPort);
            xDataSourceResH = new AttributeRessourceHandler("XDataSource", ControlPort);
            yDataSourceResH = new AttributeRessourceHandler("YDataSource", ControlPort);
            zDataSourceResH = new AttributeRessourceHandler("ZDataSource", ControlPort);
            xDataSource.IsValid = false; yDataSource.IsValid = false; zDataSource.IsValid = false;
            //dataSources.AddRange(LocalRessourceLocator.SearchRessource("ChannelOutput"));
        }

        protected override void InitializeAxes()
        {
            xAxis = new Axis(
                        new Vertex(),
                        new Vertex(Visualizer.NOMINAL_LENGTH / 2, 0, 0),
                        new Vertex(0, -1, -1),
                        View, "X");
            
            yAxis = new Axis(
                        new Vertex(0, -Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(0, Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(-1, 0, 1),
                        View, "Y");
            zAxis = new Axis(
                        new Vertex(),
                        new Vertex(0, 0, Visualizer.NOMINAL_LENGTH / 2),
                        new Vertex(-1, -1, 0),
                        View, "Z");
            AxesOrigin.AddChildren(xAxis, yAxis, zAxis);
        }

        public override void NewRessources(URL[] newRessources)
        {
            base.NewRessources(newRessources);
            for (int i = 0; i < newRessources.Length; i++)
            {
                if (newRessources[i].Ressource == "ChannelOutput") { dataSources.Add(newRessources[i]); }
            }
            dataSourcesResH.Advertise();
        }
        public override void RessourcesLost(URL[] lostRessources)
        {
            base.RessourcesLost(lostRessources);
            for (int i = 0; i < lostRessources.Length; i++) 
            {
                if (xDataSource.Equals(lostRessources[i])) 
                { 
                    xDataSource = new URL(); xDataSource.IsValid = false;
                    XDisplayMode = DisplayModes.NONE;
                }
                if (yDataSource.Equals(lostRessources[i])) 
                { 
                    yDataSource = new URL(); yDataSource.IsValid = false;
                    YDisplayMode = DisplayModes.NONE;
                }
                if (zDataSource.Equals(lostRessources[i])) 
                { 
                    zDataSource = new URL(); zDataSource.IsValid = false;
                    ZDisplayMode = DisplayModes.NONE;
                }
                dataSources.Remove(lostRessources[i]); 
            }
            dataSourcesResH.Advertise();
        }

        public List<URL> DataSources 
        { 
            get { return dataSources; } 
            protected set 
            {

            } 
        }
        public URL XDataSource 
        {
            get { return xDataSource; } 
            set 
            {
                if (xDataSource.IsValid) { xPort.Quiet(xDataSource.Device, xDataSource.Ressource, "X"); }
                xDataSource = value;
                xPort.Listen(xDataSource.Device, xDataSource.Ressource, "X");
            } 
        }
        public URL YDataSource
        {
            get { return yDataSource; }
            set
            {
                if (yDataSource.IsValid) { yPort.Quiet(yDataSource.Device, yDataSource.Ressource, "Y"); }
                yDataSource = value;
                yPort.Listen(yDataSource.Device, yDataSource.Ressource, "Y");
            }
        }
        public URL ZDataSource
        {
            get { return zDataSource; }
            set
            {
                if (zDataSource.IsValid) { zPort.Quiet(zDataSource.Device, zDataSource.Ressource, "Z"); }
                zDataSource = value;
                zPort.Listen(zDataSource.Device, zDataSource.Ressource, "Z");
            }
        }

        public virtual DataSet X
        {
            get { return currentX; }
            set { currentX = value; if (!freezed) { ForceDisplayUpdate(); } }
        }
        public virtual DataSet Y
        {
            get { return currentY; }
            set { currentY = value; if (!freezed) { ForceDisplayUpdate(); } }
        }
        public virtual DataSet Z
        {
            get { return currentZ; }
            set { currentZ = value; if (!freezed) { ForceDisplayUpdate(); } }
        }

        public virtual float XDisplacement
        {
            get { return xAxis.Displacement; }
            set { xAxis.Displacement = value; xDisplacementResH.Advertise(value); }
        }
        public virtual float YDisplacement
        {
            get { return yAxis.Displacement; }
            set { yAxis.Displacement = value; yDisplacementResH.Advertise(value); }
        }
        public virtual float ZDisplacement
        {
            get { return zAxis.Displacement; }
            set
            { zAxis.Displacement = value; zDisplacementResH.Advertise(value); }
        }

        public virtual float XScaling
        {
            get { return xAxis.Scaling; }
            set { xAxis.Scaling = value; xScalingResH.Advertise(value); }
        }
        public virtual float YScaling
        {
            get { return yAxis.Scaling; }
            set { yAxis.Scaling = value; yScalingResH.Advertise(value); }
        }
        public virtual float ZScaling
        {
            get { return zAxis.Scaling; }
            set { zAxis.Scaling = value; zScalingResH.Advertise(value); }
        }

        public virtual DisplayModes XDisplayMode
        {
            get { return xAxisDisplayMode; }
            set
            {
                Axis axis;
                
                xAxisDisplayMode = value;
               
                switch (xAxisDisplayMode)
                {
                    case DisplayModes.HISTORY:
                        axis = new HistoryAxis(
                        new Vertex(),
                        new Vertex(Visualizer.NOMINAL_LENGTH / 2, 0, 0),
                        new Vertex(0, -1, -1),
                        View, "X");
                        break;
                    default:
                        axis = new Axis(
                        new Vertex(),
                        new Vertex(Visualizer.NOMINAL_LENGTH / 2, 0, 0),
                        new Vertex(0, -1, -1),
                        View, "X");
                        break;
                    case DisplayModes.SAMPLE_SLAVE:
                        axis = new DataSetSlaveAxis(
                        new Vertex(),
                        new Vertex(Visualizer.NOMINAL_LENGTH / 2, 0, 0),
                        new Vertex(0, -1, -1),
                        View, "X");
                        break;
                    case DisplayModes.SAMPLE_MASTER:
                        axis = new DataSetMasterAxis(
                        new Vertex(),
                        new Vertex(Visualizer.NOMINAL_LENGTH / 2, 0, 0),
                        new Vertex(0, -1, -1),
                        View, "X");
                        break;
                    case DisplayModes.INCREMENT:
                        axis = new IncrementAxis(
                        new Vertex(),
                        new Vertex(Visualizer.NOMINAL_LENGTH / 2, 0, 0),
                        new Vertex(0, -1, -1),
                        View, "X");
                        break;
                }
                lock (displayUpdateLock)
                {
                    if (xAxis != null) { xAxis.Dispose(); }
                    xAxis = axis;
                    AxesOrigin.AddChild(xAxis);
                }
                SetDisplayType(xAxis.Dimensions + yAxis.Dimensions + zAxis.Dimensions);
            }
        }
        public virtual DisplayModes YDisplayMode
        {
            get { return yAxisDisplayMode; }
            set
            {
                Axis axis;
                
                yAxisDisplayMode = value;
                switch (yAxisDisplayMode)
                {
                    case DisplayModes.HISTORY:
                        axis = new HistoryAxis(
                        new Vertex(0, -Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(0, Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(-1, 0, 1),
                        View, "Y");
                        break;
                    default:
                        axis = new Axis(
                        new Vertex(0, -Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(0, Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(-1, 0, 1),
                        View, "Y");
                        break;
                    case DisplayModes.SAMPLE_SLAVE:
                        axis = new DataSetSlaveAxis(
                        new Vertex(0, -Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(0, Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(-1, 0, 1),
                        View, "Y");
                        break;
                    case DisplayModes.SAMPLE_MASTER:
                        axis = new DataSetMasterAxis(
                        new Vertex(0, -Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(0, Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(-1, 0, 1),
                        View, "Y");
                        break;
                    case DisplayModes.INCREMENT:
                        axis = new IncrementAxis(
                        new Vertex(0, -Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(0, Visualizer.NOMINAL_LENGTH / 2, 0),
                        new Vertex(-1, 0, 1),
                        View, "Y");
                        break;
                }
                lock (displayUpdateLock)
                {
                    if (yAxis != null) { yAxis.Dispose(); }

                    yAxis = axis;
                    AxesOrigin.AddChild(yAxis);
                }

                SetDisplayType(xAxis.Dimensions + yAxis.Dimensions + zAxis.Dimensions);
            }
        }
        public virtual DisplayModes ZDisplayMode
        {
            get { return zAxisDisplayMode; }
            set
            {
                Axis axis;
                
                zAxisDisplayMode = value;                
                switch (zAxisDisplayMode)
                {
                    case DisplayModes.HISTORY:
                        axis = new HistoryAxis(
                        new Vertex(),
                        new Vertex(0, 0, Visualizer.NOMINAL_LENGTH / 2),
                        new Vertex(-1, -1, 0),
                        View, "Z");
                        break;
                    default:
                        axis = new Axis(
                        new Vertex(),
                        new Vertex(0, 0, Visualizer.NOMINAL_LENGTH / 2),
                        new Vertex(-1, -1, 0),
                        View, "Z");
                        break;
                    case DisplayModes.SAMPLE_SLAVE:
                        axis = new DataSetSlaveAxis(
                        new Vertex(),
                        new Vertex(0, 0, Visualizer.NOMINAL_LENGTH / 2),
                        new Vertex(-1, -1, 0),
                        View, "Z");
                        break;
                    case DisplayModes.SAMPLE_MASTER:
                        axis = new DataSetMasterAxis(
                        new Vertex(),
                        new Vertex(0, 0, Visualizer.NOMINAL_LENGTH / 2),
                        new Vertex(-1, -1, 0),
                        View, "Z");
                        break;
                    case DisplayModes.INCREMENT:
                        axis = new IncrementAxis(
                        new Vertex(),
                        new Vertex(0, 0, Visualizer.NOMINAL_LENGTH / 2),
                        new Vertex(-1, -1, 0),
                        View, "Z");
                        break;
                }

                lock (displayUpdateLock)
                {
                    if (zAxis != null) { zAxis.Dispose(); }
                    zAxis = axis;
                    AxesOrigin.AddChild(zAxis);
                }
                SetDisplayType(xAxis.Dimensions + yAxis.Dimensions + zAxis.Dimensions);
            }
        }



        protected override float[][][] GetDisplayData(int position)
        {
            int valuesLength;
            DataSet x, y, z;

            lock (displayUpdateLock)
            {
                if (!freezed)
                {
                    valuesLength = int.MaxValue;
                    x = currentX;
                    y = currentY;
                    z = currentZ;

                    if (x != null && x.BlocksLength < valuesLength) { valuesLength = x.BlocksLength; }
                    if (y != null && y.BlocksLength < valuesLength) { valuesLength = y.BlocksLength; }
                    if (z != null && z.BlocksLength < valuesLength) { valuesLength = z.BlocksLength; }
                    if (valuesLength == int.MaxValue) { valuesLength = 100; }

                    xAxis.CurrentDataSet = x;
                    yAxis.CurrentDataSet = y;
                    zAxis.CurrentDataSet = z;

                    xAxis.CreateValues(valuesLength);
                    yAxis.CreateValues(valuesLength);
                    zAxis.CreateValues(valuesLength);
                }
                else
                {
                    if (position > xAxis.HistoryCount) { position = xAxis.HistoryCount - 1; }
                    if (position > yAxis.HistoryCount) { position = yAxis.HistoryCount - 1; }
                    if (position > zAxis.HistoryCount) { position = zAxis.HistoryCount - 1; }
                }
            }

            return new float[][][] { xAxis.GetValues(position), yAxis.GetValues(position), zAxis.GetValues(position) };
            
        }

    }
}
