using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Math32.LinearAlgebra
{
    public struct Vertex
    {
        public float x, y, z;
        public int color;

        public Vertex(float x, float y, float z, Color color)
        {
            this.x = x; this.y = y; this.z = z; this.color = color.ToArgb();
        }
        public Vertex(float x, float y, float z, int color)
        {
            this.x = x; this.y = y; this.z = z; this.color = color;
        }
        public Vertex(float x, float y, float z) { this.x = x; this.y = y; this.z = z; color = 0; }

        public void Transform(Matrix44 m)
        {
            float tX = x * m[0, 0] + y * m[1, 0] + z * m[2, 0] + m[3, 0];
            float tY = x * m[0, 1] + y * m[1, 1] + z * m[2, 1] + m[3, 1];
            float tZ = x * m[0, 2] + y * m[1, 2] + z * m[2, 2] + m[3, 2];
            x = tX; y = tY; z = tZ;
        }

        public Vertex CrossProduct(Vertex v) { return Vertex.CrossProduct(this, v); }
        public static Vertex CrossProduct(Vertex a, Vertex b)
        {
            Vertex c = new Vertex();
            c.x = a.y * b.z - a.z * b.y;
            c.y = -(a.x * b.z - a.z * b.x);
            c.z = a.x * b.y - a.y * b.x;
            return c;
        }

        public float DotProduct(Vertex v) { return Vertex.DotProduct(this, v); }
        public static float DotProduct(Vertex a, Vertex b) { return a.x * b.x + a.y * b.y + a.z * b.z;}

        public void Normalize() { this = Vertex.Normalize(this); }
        public static Vertex Normalize(Vertex vertex)
        {
            float length = Magnitude(vertex);
            if (length == 0) { return vertex; }
            float olength = 1.0f / length;
            vertex.x *= olength;
            vertex.y *= olength;
            vertex.z *= olength;

            return vertex;
        }

        public float Magnitude() { return Vertex.Magnitude(this); }
        public static float Magnitude(Vertex a) { return (float)Math.Sqrt((a.x * a.x) + (a.y * a.y) + (a.z * a.z)); }

        public static Vertex operator -(Vertex v1, Vertex v2)
        {
            Vertex ret = new Vertex();
            ret.x = v1.x - v2.x;
            ret.y = v1.y - v2.y;
            ret.z = v1.z - v2.z;
            return ret;
        }

        public static Vertex operator +(Vertex v1, Vertex v2)
        {
            Vertex ret = new Vertex();
            ret.x = v1.x + v2.x;
            ret.y = v1.y + v2.y;
            ret.z = v1.z + v2.z;
            return ret;
        }
        public static float operator *(Vertex v1, Vertex v2) { return DotProduct(v1, v2); }
        public static Vertex operator +(Vertex v, float f)
        {
            v.Transform(Matrix44.Scaling(1.0f + f / Magnitude(v)));
            return v;
        }
        public static Vertex operator -(Vertex v, float f)
        {
            v.Transform(Matrix44.Scaling(1.0f - f / Magnitude(v)));
            return v;
        }
        public static Vertex operator *(Vertex v, float f)
        {
            v.Transform(Matrix44.Scaling( f * Magnitude(v)));
            return v;
        }
        public static Vertex operator /(Vertex v, float f)
        {      
            v.Transform(Matrix44.Scaling( 1/f ));
            return v;
        }

        public override bool Equals(object obj)
        {
            if (obj is Vertex)
            {
                Vertex v = (Vertex)obj;

                if (x == v.x && y == v.y && z == v.z && color == v.color) { return true; }
                return false;
            }

            return false;
        }
        public override string ToString() { return x + ", " + y + ", " + z; }
    }

    

}
