using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using Motr3D.Primitives;
using Math32.LinearAlgebra;

namespace Motr3D
{
    public abstract class Motr3DImplementation: IMotr3DImplementation
    {
        protected Dictionary<Object3D, IObject3DImplementation> implementations =
            new Dictionary<Object3D, IObject3DImplementation>(200);
        private bool disposed = false;
        protected Control renderingControl, viewport;
        protected int renderHeight = 200, renderWidth = 200;

        public event EventHandler DeviceReset;
        public event EventHandler DeviceReady;

        public Motr3DImplementation() 
        { 
            //viewport.Dock = DockStyle.Fill;
        }

        public virtual Control ViewPort { get { return viewport; } }
        public virtual int RenderHeight { get { return renderHeight; } set { renderHeight = value; ResetRenderingEngine(); } }
        public virtual int RenderWidth { get { return renderWidth; } set { renderWidth = value; ResetRenderingEngine(); } }
        public abstract FillMode CurrentFillMode { get; set; }
        public abstract CullMode CurrentCullMode { get; set; }
        public abstract Lighting EnableLighting { get; set; }
        public abstract bool DepthBuffer { get; set; }
        public bool Disposed { get { return disposed; } }
        public  void Dispose() { Dispose(true); }
        protected virtual void Dispose(bool managedRessources)
        {
            if (!disposed)
            {
                if (managedRessources)
                {
                    DisposeRenderingEngine();
                    foreach (IObject3DImplementation imp in implementations.Values) { imp.Dispose(); }
                    GC.SuppressFinalize(this);
                }
            }
            else { throw new ObjectDisposedException(this.ToString(), "This object has been disposed previously"); }
            disposed = true;
        }
        protected abstract void DisposeRenderingEngine();

        public abstract void InitializeImplementation(Control viewport);
        public abstract void ResetRenderingEngine();
        protected object drawLock = new object();

        public abstract void ClearTarget(Color color);
        public abstract void BeginRendering();
        public abstract void EndRendering();

        public abstract void SetPerspectiveProjection(float fieldOfView, float aspectRatio, float nearClipDistance, float farClipDistance);
        public abstract void SetOrthogonalProjection(float width, float height, float nearClipDistance, float farClipDistance);
        public abstract void TransformCamera(Matrix44 transform);
        public abstract void TransformWorld(Math32.LinearAlgebra.Matrix44 transform);
        
        public abstract Bitmap GetLastRender();
        public abstract void DisplayRender();

        public virtual void Target(Camera camera)
        {
            ICameraImplementation camIm;
            if (implementations.ContainsKey(camera)) 
            {
                camIm = (ICameraImplementation)implementations[camera]; 
            }
            else
            {
                camIm = (ICameraImplementation)MakeImplementation(camera);
                implementations.Add(camera, camIm);
                camera.Disposing += new EventHandler(ImplementationDisposed);
            }
            camIm.Target();
        }

        public void DrawObject(Object3D obj)
        {
            IObject3DImplementation objIm;
            if(implementations.ContainsKey(obj)){ objIm = implementations[obj]; }
            else 
            {
                /* Here we try to lock the Object3D obj before creating its implementation. Should the lock method
                 * timeout, this method returns and we will try creating the Object3D implementations
                 * again next time the render loop encounters obj. */
                //The timeout is 0, which mean the method returns immediatly with the result
                if (!obj.Lock(0)) { return; }
                objIm =  MakeImplementation(obj);
                try { implementations.Add(obj, objIm); }
                catch (IndexOutOfRangeException) // The Dictionnay is full !!!
                {
                    implementations = new Dictionary<Object3D, IObject3DImplementation>(implementations);
                }
                obj.Disposing += new EventHandler(ImplementationDisposed);
                obj.Retiring += new EventHandler(ImplementationDisposed);
                if (obj is Primitive)
                {
                    ((Primitive)obj).PrimitiveChanged += new EventHandler(ImplementationDisposed);
                }
                obj.Unlock(); //Release the lock on obj
            }
            objIm.DrawSelf();
        }

        protected abstract IObject3DImplementation MakeImplementation(Object3D object3D);
   
        protected void ImplementationDisposed(object sender, EventArgs args) 
        {
            Object3D obj = sender as Object3D;
            //obj.Disposing -= new EventHandler(ImplementationDisposed);
            if (!implementations.ContainsKey(obj)) { return; }
            implementations[obj].Dispose();
            implementations.Remove(obj);
            
        }

        public void OnDeviceReset() { if (DeviceReset != null) { DeviceReset(this, null); } }
        public void OnDeviceReady() { if (DeviceReady != null) { DeviceReady(this, null); } }
    }
}
