using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Motr3D.Exceptions;
using Motr3D.Primitives;
using System.Threading;
using Math32.LinearAlgebra;

namespace Motr3D
{
    public delegate void Motr3DViewportEventHandler(Motr3DViewport sender);

    public partial class Motr3DViewport : UserControl
    {
        public const Lighting DEFAULT_LIGHTING = Lighting.NO;
        public const CullMode DEFAULT_CULLMODE = CullMode.NONE;
        public const FillMode DEFAULT_FILLMODE = FillMode.WIREFRAME;
        public const Projection DEFAULT_PROJECTION = Projection.PERSPECTIVE;
        public const Visibility DEFAULT_VISIBILITY = Visibility.VISIBLE;

        
        
        private IMotr3DImplementation implementation;
        private volatile bool asynchronousRendering = true;
        private volatile int renderThreadSleepMs = 30;
        private List<Camera> cameras = new List<Camera>();
        private AutoResetEvent renderEvent = new AutoResetEvent(true);
        private Thread renderThread;

        private volatile bool countFramerate = false;
        private DateTime lastFrame = DateTime.Now;
        private volatile float currentFramerate = 0;

        Visibility currentVisibility;
        Projection currentProjection;
        Camera currentCamera;

        public event Motr3DViewportEventHandler RenderStarted;
        public event Motr3DViewportEventHandler RenderEnded;
        protected Motr3DViewport() { InitializeComponent(); }
        public Motr3DViewport(IMotr3DImplementation implementation)
        {
            InitializeComponent();
            this.implementation = implementation;
            renderThread = new Thread(new ThreadStart(RenderLoop));
            renderThread.IsBackground = true;
            renderThread.Start();
        }
        public IMotr3DImplementation Implementation { get { return implementation; } }
        public void AddCameras(params Camera[] cameras)
        { 
            lock (cameras) 
            { 
                this.cameras.AddRange(cameras);
                for (int i = 0; i < cameras.Length; i++) { cameras[i].Viewport = this; }
            }
        }
        public void RemoveCamera(Camera camera) { lock (cameras) { cameras.Remove(camera); camera.Viewport = null; } }

        protected override void OnInvalidated(InvalidateEventArgs e)
        {
            base.OnInvalidated(e);
            if (!asynchronousRendering) { Render(); }
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            invalidated = true;
        }
        volatile bool invalidated = true;
        public virtual void Render() { renderEvent.Set(); Thread.Sleep(0); }
        public virtual void RenderLoop()
        {
            Thread.Sleep(200);
            
            implementation.InitializeImplementation(this);
            while (true)
            {
                if (!asynchronousRendering) { renderEvent.WaitOne(); }
                
                if (invalidated)
                {
                    implementation.ResetRenderingEngine();
                    invalidated = false;
                }

                OnRenderStarted();
                try
                {
                    implementation.BeginRendering();
                    lock (cameras)
                    {
                        foreach (Camera camera in cameras)
                        {
                            currentCamera = camera;

                            if (implementation.EnableLighting != DEFAULT_LIGHTING) { implementation.EnableLighting = DEFAULT_LIGHTING; }
                            if (implementation.CurrentCullMode != DEFAULT_CULLMODE) { implementation.CurrentCullMode = DEFAULT_CULLMODE; }
                            if (implementation.CurrentFillMode != DEFAULT_FILLMODE) { implementation.CurrentFillMode = DEFAULT_FILLMODE; }
                            ChangeProjection(DEFAULT_PROJECTION, camera);
                            currentProjection = DEFAULT_PROJECTION;
                            currentVisibility = DEFAULT_VISIBILITY;

                            implementation.Target(camera);
                            implementation.ClearTarget(camera.BackgroundColor);

                            DisplayCamera(camera);
                            for (int i = 0; i < camera.RenderNodes.Count; i++)
                            {
                                DisplayObject(camera.RenderNodes[i], Matrix44.Identity);
                            }
                            implementation.Target(null);
                            implementation.DrawObject(camera);


                        }
                    }
                    implementation.EndRendering();
                    implementation.DisplayRender();
                }
                catch (OutOfVideoMemoryException e) { }
                catch (Motr3DException e) { }

                if (countFramerate) { currentFramerate = (currentFramerate + (1 / (DateTime.Now.Ticks - lastFrame.Ticks))) / 2; }

                OnRenderEnded();
                Thread.Sleep(renderThreadSleepMs);
            }
        }


        private void DisplayObject(Object3D obj, Matrix44 transform)
        {
            Visibility visibility;
            if (!obj.Lock(20)) { return; }

            implementation.EnableLighting = obj.Lighting;
            implementation.CurrentCullMode = obj.CullMode;
            implementation.CurrentFillMode = obj.FillMode;
            ChangeProjection(obj.Projection, currentCamera);
            if (obj.Visibility != Visibility.CURRENT)
            {
                currentVisibility = obj.Visibility;
                visibility = obj.Visibility;
            }
            else { visibility = currentVisibility; }
                        
            if (obj is Text2D)
            {
                transform = obj.currentTranslationMatrix * transform;
                Matrix44 m = Matrix44.Identity;
                m[0, 0] = lac[0, 0]; m[0, 1] = lac[1, 0]; m[0, 2] = lac[2, 0];
                m[1, 0] = lac[0, 1]; m[1, 1] = lac[1, 1]; m[1, 2] = lac[2, 1];
                m[2, 0] = lac[0, 2]; m[2, 1] = lac[1, 2]; m[2, 2] = lac[2, 2];
                m = obj.WorldScalingMatrix * obj.currentRotationMatrix * m;
                m[0, 3] = transform[0, 3]; m[1, 3] = transform[1, 3]; m[2, 3] = transform[2, 3];
                m[3, 0] = transform[3, 0]; m[3, 1] = transform[3, 1]; m[3, 2] = transform[3, 2]; m[3, 3] = transform[3, 3];
                transform = m;
            }
            else
            {
                transform =
                obj.currentScalingMatrix *
                obj.currentRotationMatrix *
                obj.currentTranslationMatrix *
                transform;
            }
            if (currentVisibility != Visibility.HIDDEN)
            {
                if (!(obj is Camera)) { implementation.TransformWorld(transform); }
                if (obj is Primitive) 
                { 
                   if (((Primitive)obj).Transparency < byte.MaxValue) { implementation.DepthBuffer = false; } 
                }
                //implementation.DrawObject(obj);
                obj.Draw(currentCamera);

                implementation.DepthBuffer = true;
            }
            for (int i = 0; i < obj.Children.Count; i++)
            {
                if (!obj.Children[i].BlockRenderLoop) { DisplayObject(obj.Children[i], transform); }
                implementation.EnableLighting = obj.Lighting;
                implementation.CurrentCullMode = obj.CullMode;
                implementation.CurrentFillMode = obj.FillMode;
                ChangeProjection(obj.Projection, currentCamera);
                currentVisibility = visibility;
            }
            obj.Unlock();
        }

        public Matrix44 lac = Matrix44.Identity;
        public virtual void DisplayCamera(Camera camera)
        {
            Vertex norm = new Vertex(1, 0, 0);//objectNormal;

            norm.Transform(camera.WorldTransformMatrix);

            lac = Matrix44.LookAtCamera(camera.PositionInWorld, norm);
            implementation.TransformCamera(Matrix44.LookAtCamera(camera.PositionInWorld, norm));
        }

        protected virtual void ChangeProjection(Projection proj, Camera camera)
        {
            if (proj == Projection.PERSPECTIVE)
            {
                implementation.SetPerspectiveProjection(
                    camera.FieldOfView, camera.AspectRatio, camera.NearClipDistance, camera.FarClipDistance);
            }
            else if (proj == Projection.ORTHOGONAL)
            {
                implementation.SetOrthogonalProjection(
                    camera.Width, camera.Height, camera.NearClipDistance, camera.FarClipDistance);
            }
        }

        public void OnRenderStarted() { if (RenderStarted != null) { RenderStarted(this); } }
        public void OnRenderEnded() { if (RenderEnded != null) { RenderEnded(this); } }
    }
}
