using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Motr3D.Primitives;
using System.Threading;
using Math32.LinearAlgebra;
using Motr3D.Prebuilt;

namespace Motr3D
{
    public enum Projection { ORTHOGONAL = 4, PERSPECTIVE = 2, CURRENT = 0 };
    public enum Visibility { HIDDEN = 4, VISIBLE = 2, CURRENT = 0 };
    public enum FillMode { SOLID = 8, WIREFRAME = 4, POINT = 2, CURRENT = 0 };
    public enum CullMode { COUNTERCLOCKWISE = 8, CLOCKWISE = 4, NONE = 2, CURRENT = 0 };
    public enum Lighting { YES = 4, NO = 2, CURRENT = 0 };

    public delegate void TranslatedEventHandler(Object3D sender, Vertex translation);
    public delegate void RotatedEventHandler(Object3D sender, float rX, float rY, float rZ);
    public delegate void ScaledEventHandler(Object3D sender, float sX, float sY, float sZ);
    public delegate void PointedAtEventHandler(Object3D sender, Vertex target);
    public delegate void DrawingEventHandler(Object3D sender, Camera camera);
    public delegate void Object3DEventHandler(Object3D sender);
    public delegate void Object3DLockEventHandler(Object3D sender, bool locked);

    public abstract class Object3D: IDisposable
    {
        public const int OBJECT_DRAW_TIMEOUT_MS = 500;
        
        protected static God god = new God();

        private object tag;

        protected string name = "Object 3D";
        protected List<Object3D>children = new List<Object3D>();
        protected Object3D parent = god;
        protected Matrix44 currentTransform = Matrix44.Identity;
        protected Matrix44 currentWorldTransform = Matrix44.Identity;
        private bool showAxes = false; private Axes axes;
        private volatile bool blockRenderLoop = false;
        private bool disposed = false;
        protected object object3DLock = new object();
        protected FillMode fillMode = FillMode.CURRENT;
        protected CullMode cullMode = CullMode.CURRENT;
        protected Lighting lighting = Lighting.CURRENT;
        protected Visibility visibility = Visibility.CURRENT;
        protected Projection projection = Projection.CURRENT;
        protected bool locked = false;

        public event EventHandler Disposing;
        public event EventHandler Retiring;
        public event TranslatedEventHandler Translated;
        public event RotatedEventHandler Rotated;
        public event ScaledEventHandler Scaled;
        public event PointedAtEventHandler PointedAt;
        public event Object3DEventHandler VisibilityChanged;
        public event Object3DEventHandler LightingChanged;
        public event Object3DEventHandler CullModeChanged;
        public event Object3DEventHandler FillModeChanged;
        public event Object3DEventHandler ProjectionChanged;
        public event Object3DEventHandler ParentChanged;
        public event Object3DEventHandler ChildrenChanged;
        public event DrawingEventHandler Drawing;
        public event Object3DLockEventHandler Locked;

        ~Object3D() { Dispose(false); }

        public void Dispose() { Dispose(true); }
        protected virtual void Dispose(bool dispose)
        {
            Monitor.Enter(this);
            if (dispose)
            {
                //Dispose managed ressources
                while (children.Count != 0) { children[0].Dispose(); }
                parent.RemoveChild(this);
                if (parent != null) { }
                GC.SuppressFinalize(this);
                disposed = true;
                OnDisposing();
            }
            Monitor.Exit(this);
        }

        public virtual string Name { get { return name; } set { name = value; } }
        public bool Disposed { get { return disposed; } }
        public virtual FillMode FillMode { get { return fillMode; } set { fillMode = value; OnFillModeChanged(); } }
        public virtual CullMode CullMode { get { return cullMode; } set { cullMode = value; OnCullModeChanged(); } }
        public virtual Lighting Lighting { get { return lighting; } set { lighting = value; OnLightingChanged(); } }
        public virtual Visibility Visibility { get { return visibility; } set { visibility = value; OnVisibilityChanged(); } }
        public virtual Projection Projection { get { return projection; } set { projection = value; OnProjectionChanged(); } }
        public virtual bool BlockRenderLoop 
        { 
            get { return blockRenderLoop; } 
            set 
            { 
                blockRenderLoop = value;
                OnRetiring();
                lock (object3DLock) { foreach (Object3D o in children) { o.OnRetiring(); } }
            } 
        }
        /* An Object3D could have many parents, which would duplicate its rendering but the object 
         would still uses the same ressources*/
        public Object3D Parent 
        { 
            get { return parent; } 
            set 
            {
                lock (object3DLock) { parent = value; if (value == null) { parent = god; } OnParentChanged(); }
            } 
        }
        public List<Object3D> Children
        {
            get { return children; }
            set { lock (object3DLock) { children = value; } } 
        }
       
        
        public virtual Matrix44 WorldTranslationMatrix 
        { 
            get { return parent.WorldTranslationMatrix * currentTranslationMatrix; } 
        }
        public virtual Matrix44 WorldRotationMatrix
        {
            get { return parent.WorldRotationMatrix * currentRotationMatrix;}
        }
        public virtual Matrix44 WorldScalingMatrix
        {
            get { return parent.WorldScalingMatrix * currentScalingMatrix; }
        }
        public virtual Matrix44 WorldTransformMatrix
        {
            get { return currentRotationMatrix * currentScalingMatrix * currentTranslationMatrix * parent.WorldTransformMatrix; }
        }
        public virtual Vertex PositionInWorld
        {
            get 
            { 
                Vertex pos = new Vertex();
                pos.Transform( WorldTransformMatrix );
                return pos; 
            }
            set { MoveTo(value); }
        }

        public object Tag { get { return tag; } set { tag = value; } }

        public virtual void AddChild(Object3D child) 
        {
            lock (object3DLock) { children.Add(child); child.Parent = this; } 
            OnChildrenChanged();
        }
        public virtual void AddChildren(params Object3D[] children) 
        {
            lock (object3DLock) { for (int i = 0; i < children.Length; i++) { AddChild(children[i]); } }
        }
        public virtual void AddChildren(List<Object3D> children) 
        {
            lock (object3DLock) { foreach (Object3D child in children) { AddChild(child); } }
        }

        public virtual void RemoveChild(Object3D child) { lock (object3DLock) { children.Remove(child); child.Parent = god; } }
        public virtual void RemoveChildren()
        {
            lock (object3DLock) { foreach (Object3D o in children) { o.Parent = god; } children.Clear(); }
            OnChildrenChanged();
        }
        public virtual Matrix44 CurrentTransform { get { return currentTransform; } set { currentTransform = value; } }
        public virtual bool ShowAxes
        {
            get { return showAxes; }
            set
            {
                if (value == true) { axes = new Axes(); AddChild(axes); }
                else { axes.Dispose(); }
                showAxes = value;
            }
        }

        //public virtual void Translate(VertexStruct vertex) { PropagateTranslation(vertex.x, vertex.y, vertex.z); }
        public virtual void Translate(Vertex vertex) { currentTranslationMatrix.Translate(vertex.x, vertex.y, vertex.z); }
        //public virtual void Translate(float dx, float dy, float dz) { PropagateTranslation(dx, dy, dz); }
        public virtual void Translate(float dx, float dy, float dz) { currentTranslationMatrix.Translate(dx, dy, dz); }
        //maybe MoveTo is broken
        public virtual void MoveTo(Vertex vertex) { Translate(vertex-PositionInWorld); }
        public virtual void MoveTo(float x, float y, float z) { Translate(new Vertex(x, y, z) - PositionInWorld); }
        public virtual void ResetPosition() { currentTranslationMatrix = Matrix44.Identity; }

        public virtual void DirectionnalTranslate(Vertex vertex)
        {
            Vertex norm = vertex;
            norm.Transform(currentRotationMatrix);
            Translate(norm);
        }
        public virtual void DirectionnalTranslate(float dX, float dY, float dZ)
        {
            DirectionnalTranslate(new Vertex(dX, dY, dZ));
        }

        #region Rotation Methods
            public virtual void Rotate(float roll, float pitch, float yaw) 
            {
                rotationZ += roll;
                rotationX += pitch;
                rotationY += yaw;
                currentRotationMatrix = Matrix44.Rotation(rotationZ, rotationX, rotationY);
                OnRotated(roll, yaw, pitch);
            }
            protected float rotationX = 0;
            public virtual void RotateX(float degrees) 
            { 
                //currentRotationMatrix.RotateX(degrees);
                //ResetRotation();
                rotationX += degrees;
                //currentRotationMatrix.RotateX(rotationX);
                //currentRotationMatrix.RotateY(rotationY);
                //currentRotationMatrix.RotateZ(rotationZ);
                currentRotationMatrix = Matrix44.Rotation(rotationZ, rotationX, rotationY);
                OnRotated(degrees, 0.0f, 0.0f);
            }
            protected float rotationY = 0;
            public virtual void RotateY(float degrees) 
            { 
                //currentRotationMatrix.RotateY(degrees);
                rotationY += degrees;
                /*ResetRotation();
                currentRotationMatrix.RotateX(rotationX);
                currentRotationMatrix.RotateY(rotationY);
                currentRotationMatrix.RotateZ(rotationZ);*/
                currentRotationMatrix = Matrix44.Rotation(rotationZ, rotationX, rotationY);
                OnRotated(0.0f, degrees, 0.0f);
            }
            protected float rotationZ = 0;
            public virtual void RotateZ(float degrees) 
            { 
                //currentRotationMatrix.RotateZ(degrees);
                rotationZ += degrees;
                /*ResetRotation();
                currentRotationMatrix.RotateX(rotationX);
                currentRotationMatrix.RotateY(rotationY);
                currentRotationMatrix.RotateZ(rotationZ);*/
                currentRotationMatrix = Matrix44.Rotation(rotationZ, rotationX, rotationY);
                OnRotated(0.0f, 0.0f, degrees);
            }
            public virtual void RotateVector(float degrees, Vertex vector)
            {
                currentRotationMatrix.RotateVector(degrees, vector);
            }
            public virtual void ResetRotation()
            {
                rotationX = 0; rotationY = 0; rotationZ = 0;
                currentRotationMatrix = Matrix44.Identity; 
            }
        #endregion

        #region Scaling Methods
            public virtual void Scale(float scaleFactor) 
            {
                currentScalingMatrix.Scale(scaleFactor);
                OnScaled(scaleFactor, scaleFactor, scaleFactor);
            }
            public virtual void ScaleX(float scaleFactor) 
            {
                currentScalingMatrix.ScaleX(scaleFactor); 
                OnScaled(scaleFactor, 0.0f, 0.0f); 
            }
            public virtual void ScaleY(float scaleFactor) 
            { 
                currentScalingMatrix.ScaleY(scaleFactor);
                OnScaled(0.0f, scaleFactor, 0.0f);
            }
            public virtual void ScaleZ(float scaleFactor) 
            { 
                currentScalingMatrix.ScaleZ(scaleFactor);
                OnScaled(0.0f, 0.0f, scaleFactor);
            }
        public virtual void ScaleVector(Vertex v)
        {
            currentScalingMatrix.ScaleVector(v);
            OnScaled(v.x, v.y, v.z);
        }
            public virtual void ResetScaling() { currentScalingMatrix = Matrix44.Identity; }
        #endregion

        public virtual void Target( Vertex target)
        {
            /*VertexStruct transformedObjectNormal = objectNormal;
            objectNormal.Transform(WorldTransformMatrix);
            transformedObjectNormal = VertexStruct.Normalize(transformedObjectNormal);
            target = VertexStruct.Normalize(target);
            VertexStruct rotationVector = VertexStruct.CrossProduct(transformedObjectNormal, target);
            double angle = Math.Acos(VertexStruct.DotProduct(transformedObjectNormal, target));
            angle *= 180 / Math.PI;

            RotateVector((float)angle, rotationVector);*/
            /*Matrix target = Matrix.Identity;
            target[3,0] = target.x*/
        }

        protected Vertex objectNormal = new Vertex(1.0f, 0.0f, 0.0f);
        public Matrix44 currentRotationMatrix = Matrix44.Identity;
        public Matrix44 currentTranslationMatrix = Matrix44.Identity;
        public Matrix44 currentScalingMatrix = Matrix44.Identity;
        public Matrix44 currentLookAtMatrix =
            Matrix44.LookAtCamera(new Vertex(0.0f, 0.0f, 0.0f), new Vertex(1.0f, 0.0f, 0.0f));
        public Matrix44 currentTransformationMatrix = Matrix44.Identity;
        public void Draw(Camera camera) 
        {
            if (Monitor.TryEnter(this, OBJECT_DRAW_TIMEOUT_MS))
            {
                OnDrawing(camera);
                DrawSelf(camera);
                Monitor.Exit(this);
            }
        }
        protected abstract void DrawSelf(Camera camera);

        public bool Lock() { Monitor.Enter(this); OnLocked(true); return true; }
        public bool Lock(int msTimeout) 
        {
            if (Monitor.TryEnter(this, msTimeout))
            {
                OnLocked(true);
                return true;
            }
            return false; 
        }
        public void Unlock() { Monitor.Exit(this); }

        public virtual void OnTranslated(Vertex translation) { if (Translated != null) { Translated(this, translation); } }
        public virtual void OnRotated(float rX, float rY, float rZ) { if (Rotated != null) { Rotated(this, rX, rY, rZ); } }
        public virtual void OnScaled(float sX, float sY, float sZ) { if (Scaled != null) { Scaled(this, sX, sY, sZ); } }
        public virtual void OnPointedAt(Vertex target) { if (PointedAt != null) { PointedAt(this, target); } }
        public virtual void OnVisibilityChanged() { if (VisibilityChanged != null) { VisibilityChanged(this); } }
        public virtual void OnLightingChanged() { if (LightingChanged != null) { LightingChanged(this); } }
        public virtual void OnCullModeChanged() { if (CullModeChanged != null) { CullModeChanged(this); } }
        public virtual void OnProjectionChanged() { if (ProjectionChanged != null) { ProjectionChanged(this); } }
        public virtual void OnFillModeChanged() { if (FillModeChanged != null) { FillModeChanged(this); } }
        public virtual void OnParentChanged() { if (ParentChanged != null) { ParentChanged(this); } }
        public virtual void OnChildrenChanged() { if (ChildrenChanged != null) { ChildrenChanged(this); } }
        public virtual void OnDrawing(Camera camera) { if (Drawing != null) { Drawing(this, camera); } }
        public virtual void OnLocked(bool locked) { if (Locked != null) { Locked(this, locked); } }
        private void OnDisposing() { if (Disposing != null) { Disposing(this, null); } }
        protected virtual void OnRetiring() { if (Retiring != null) { Retiring(this, null); } }
    }
}
