using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Motr3D
{
    public struct VertexStruct
    {
        public float x, y, z;
        public int color;

        public VertexStruct(float x, float y, float z, Color color)
        {
            this.x = x; this.y = y; this.z = z; this.color = color.ToArgb();
        }
        public VertexStruct(float x, float y, float z, int color)
        {
            this.x = x; this.y = y; this.z = z; this.color = color;
        }
        public VertexStruct(float x, float y, float z) { this.x = x; this.y = y; this.z = z; color = 0; }

        public void Transform(Matrix m)
        {
            /*float tX, tY, tZ;
            tX = x * m[0, 0] + y * m[0, 1] + z * m[0, 2] + m[0, 3];
            tY = x * m[1, 0] + y * m[1, 1] + z * m[1, 2] + m[1, 3];
            tZ = x * m[2, 0] + y * m[2, 1] + z * m[2, 2] + m[2, 3];
            x = tX; y = tY; z = tZ;*/

            float tX, tY, tZ;
            tX = x * m[0, 0] + y * m[1, 0] + z * m[2, 0] + m[3, 0];
            tY = x * m[0, 1] + y * m[1, 1] + z * m[2, 1] + m[3, 1];
            tZ = x * m[0, 2] + y * m[1, 2] + z * m[2, 2] + m[3, 2];
            x = tX; y = tY; z = tZ;
        }

        public static VertexStruct CrossProduct(VertexStruct a, VertexStruct b)
        {
            VertexStruct c = new VertexStruct();
            c.x = a.y * b.z - a.z * b.y;
            c.y = -(a.x * b.z - a.z * b.x);
            c.z = a.x * b.y - a.y * b.x;
            return c;
        }

        public static float DotProduct(VertexStruct a, VertexStruct b)
        {
            return a.x * b.x + a.y * b.y + a.z * b.z;
        }

        public static VertexStruct Normalize(VertexStruct a)
        {
            float length = Magnitude(a);
            if (length == 0) { return a; }
            float olength = 1.0f / length;
            a.x *= olength;
            a.y *= olength;
            a.z *= olength;

            return a;
        }
        

        public static float Magnitude(VertexStruct a) { return (float)Math.Sqrt((a.x * a.x) + (a.y * a.y) + (a.z * a.z)); }

        public static VertexStruct operator -(VertexStruct v1, VertexStruct v2)
        {
            VertexStruct ret = new VertexStruct();
            ret.x = v1.x - v2.x;
            ret.y = v1.y - v2.y;
            ret.z = v1.z - v2.z;
            return ret;
        }

        public static VertexStruct operator +(VertexStruct v1, VertexStruct v2)
        {
            VertexStruct ret = new VertexStruct();
            ret.x = v1.x + v2.x;
            ret.y = v1.y + v2.y;
            ret.z = v1.z + v2.z;
            return ret;
        }
        public static float operator *(VertexStruct v1, VertexStruct v2) { return DotProduct(v1, v2); }
        public static VertexStruct operator +(VertexStruct v, float f)
        {
            v.Transform(Matrix.Scaling(1.0f + f / Magnitude(v)));
            return v;
        }
        public static VertexStruct operator -(VertexStruct v, float f)
        {
            v.Transform(Matrix.Scaling(1.0f - f / Magnitude(v)));
            return v;
        }
        public static VertexStruct operator *(VertexStruct v, float f)
        {
            v.Transform(Matrix.Scaling( f * Magnitude(v)));
            return v;
        }
        public static VertexStruct operator /(VertexStruct v, float f)
        {      
            v.Transform(Matrix.Scaling( 1/f ));
            return v;
        }

        public override bool Equals(object obj)
        {
            VertexStruct v = (VertexStruct)obj;

            if (x == v.x && y == v.y && z == v.z && color == v.color) { return true; }
            return false;
        }
    }

    public struct Matrix //Could be optimized using unsafe
    {
        public float[,] m;

        public static Matrix Identity
        {
            get
            {
                Matrix i;
                i.m = new float[4, 4] { { 1, 0, 0, 0 }, { 0, 1, 0, 0 }, { 0, 0, 1, 0  }, { 0, 0, 0, 1 } };
                return i;
            }
        }

        public float this[int i, int j] { get { return m[i,j]; } set { m[i,j] = value; } }


        public void Transform(Matrix transform) { this = transform * this; }
        public void Translate(float dX, float dY, float dZ) { this = Matrix.Translation(dX, dY, dZ) * this; }
        public static Matrix Translation(float dX, float dY, float dZ)
        {
            Matrix ret;
            ret.m = new float[4, 4] { { 1, 0, 0, 0 }, { 0, 1, 0, 0 }, { 0, 0, 1, 0 }, { dX, dY, dZ, 1 } };
            return ret;
        }

        public void ScaleX(float scaleFactor) { this = Matrix.XScaling(scaleFactor) * this; }
        public static Matrix XScaling(float scaleFactor)
        {
            Matrix ret;
            ret.m = new float[4, 4] { { scaleFactor, 0, 0, 0 },
                                      { 0, 1, 0, 0 }, 
                                      { 0, 0, 1, 0 }, 
                                      {0, 0, 0, 1} };
            return ret;
        }
        public void ScaleY(float scaleFactor) { this = Matrix.YScaling(scaleFactor) * this; }
        public static Matrix YScaling(float scaleFactor)
        {
            Matrix ret;
            ret.m = new float[4, 4] { { 1, 0, 0, 0 },
                                      { 0, scaleFactor, 0, 0 }, 
                                      { 0, 0, 1, 0 }, 
                                      {0, 0, 0, 1} };
            return ret;
        }
        public void ScaleZ(float scaleFactor) { this = Matrix.ZScaling(scaleFactor) * this; }
        public static Matrix ZScaling(float scaleFactor)
        {
            Matrix ret;
            ret.m = new float[4, 4] { { 1, 0, 0, 0 },
                                      { 0, 1, 0, 0 }, 
                                      { 0, 0, scaleFactor, 0 }, 
                                      {0, 0, 0, 1} };
            return ret;
        }
        public void Scale(float scaleFactor) { this = Matrix.Scaling(scaleFactor) * this; }
        public static Matrix Scaling(float scaleFactor)
        {
            Matrix ret;
            ret.m = new float[4, 4] { { scaleFactor, 0, 0, 0 },
                                      { 0, scaleFactor, 0, 0 }, 
                                      { 0, 0, scaleFactor, 0 }, 
                                      {0, 0, 0, 1} };
            return ret;
        }
        public void ScaleVector(VertexStruct v) { this = Matrix.VectorScaling(v) * this; }
        public static Matrix VectorScaling(VertexStruct v)
        {
            Matrix ret;
            ret.m = new float[4, 4] { { v.x, 0, 0, 0 },
                                      { 0, v.y, 0, 0 }, 
                                      { 0, 0, v.z, 0 }, 
                                      {0, 0, 0, 1} };
            return ret;
        }


        //public static Matrix Rotation() { }
        public void PointAt(VertexStruct cameraPostion, VertexStruct pointAt) { this *= LookAtCamera(cameraPostion, pointAt); }
        public static Matrix LookAtCamera(VertexStruct cameraPosition, VertexStruct pointAt)
        {
            /*VertexStruct s, u, upV;
            Matrix ret;

            upV.x = 0f;
            upV.y = 1f;
            upV.z = 0f;
            
            s = new VertexStruct();
            upV = new VertexStruct();
            s = pointAt - cameraPostion;
            s.y = pointAt.y - cameraPostion.y;
            s.z = pointAt.z - cameraPostion.z;
            upV.x = 0f;
            upV.y = 1f;
            upV.z = 0f;

            u = VertexStruct.CrossProduct(upV, s);
            upV = VertexStruct.CrossProduct(s, u);
            VertexStruct.Normalize(s);
            VertexStruct.Normalize(upV);

            ret.m = new float[4, 4] {{u.x, u.y, u.z, 0f},
                                    {upV.x, upV.y, upV.z, 0f},
                                    {s.x, s.y, s.z, 0f},
                                    {cameraPostion.x, cameraPostion.y, cameraPostion.z, 1.0f}};
            return ret;*/

            VertexStruct x, y, z, upV;
            Matrix ret;

            upV = new VertexStruct();
            upV.x = 0f;
            upV.y = 1f;
            upV.z = 0f;

            z = VertexStruct.Normalize(pointAt - cameraPosition);
            x = VertexStruct.Normalize(VertexStruct.CrossProduct(upV,z));
            y = VertexStruct.CrossProduct(z, x);



            ret.m = new float[4, 4] {{x.x, y.x, z.x, 0f},
                                    {x.y, y.y, z.y, 0f},
                                    {x.z, y.z, z.z, 0f},
                                    {-VertexStruct.DotProduct(x,cameraPosition), 
                                        -VertexStruct.DotProduct(y,cameraPosition), 
                                        -VertexStruct.DotProduct(z,cameraPosition), 1.0f}};
            return ret;



        }

        public void Rotate(float roll, float pitch, float yaw) { this = Matrix.Rotation(roll, pitch, yaw) * this; }
        public static Matrix Rotation(float roll, float pitch, float yaw)
        {
            return RotationY(yaw)*Matrix.RotationZ(roll) * Matrix.RotationX(pitch);
        }

        public void RotateX(float degrees) { this = Matrix.RotationX(degrees) * this; }
        public static Matrix RotationX(float degrees)
        {
            Matrix ret;
            ret.m = new float[4, 4] {{1,0,0,0},
                                    {0,(float)Math.Cos(degrees*(Math.PI/180)), -(float)Math.Sin(degrees*(Math.PI/180)),0},
                                    {0,(float)Math.Sin(degrees*(Math.PI/180)), (float)Math.Cos(degrees*(Math.PI/180)), 0},
                                    {0,0,0,1}};
            return ret;
        }

        public void RotateY(float degrees) { this = Matrix.RotationY(degrees) * this; }
        public static Matrix RotationY(float degrees)
        {
            Matrix ret;
            ret.m = new float[4, 4] {{(float)Math.Cos(degrees*(Math.PI/180)), 0, (float)Math.Sin(degrees*(Math.PI/180)), 0},
                                    {0, 1, 0, 0},
                                    {-(float)Math.Sin(degrees*(Math.PI/180)), 0, (float)Math.Cos(degrees*(Math.PI/180)), 0},
                                    {0, 0, 0, 1}};
            return ret;
        }

        public void RotateZ(float degrees) { this = Matrix.RotationZ(degrees) * this; }
        public static Matrix RotationZ(float degrees)
        {
            Matrix ret;
            ret.m = new float[4, 4] {{(float)Math.Cos(degrees*(Math.PI/180)), -(float)Math.Sin(degrees*(Math.PI/180)), 0, 0},
                                    {(float)Math.Sin(degrees*(Math.PI/180)), (float)Math.Cos(degrees*(Math.PI/180)), 0, 0},
                                    {0,0,1,0},
                                    {0,0,0,1}};
            return ret;
        }

        public void RotateVector(float degrees, VertexStruct vector) { this = Matrix.RotationVector(degrees, vector) * this; }
        public static Matrix RotationVector(float degrees, VertexStruct vector)
        {
            Matrix ret;
            float cos = (float)Math.Cos(degrees);
            float sin = (float)Math.Sin(degrees);


            ret.m = new float[4, 4] {{((vector.x*vector.x)*(1.0f-cos))+cos, 
                                     ((vector.x*vector.y)*(1.0f-cos))+(vector.z*sin),
                                     ((vector.z*vector.x)*(1.0f-cos))-(vector.y*sin),
                                     0},
                                     {((vector.x*vector.y)*(1.0f-cos))-(vector.z*sin),
                                     ((vector.y*vector.y)*(1.0f-cos))+cos,
                                     ((vector.z*vector.y)*(1.0f-cos))+(vector.x*sin),
                                     0},
                                     {((vector.x*vector.z)*(1.0f-cos))+(vector.y*sin),
                                     ((vector.y*vector.z)*(1.0f-cos))-(vector.x*sin),
                                     ((vector.z*vector.z)*(1.0f-cos))+cos,
                                     0},
                                     {0,0,0,1}};
            return ret;
        }

        public static Matrix TransposedMatrix(Matrix src)
        {
            Matrix ret = new Matrix();
            ret[0, 0] = src[0, 0];
            ret[1, 0] = src[0, 1];
            ret[2, 0] = src[0, 2];
            ret[3, 0] = 0f;
            ret[0, 1] = src[1, 0];
            ret[1, 1] = src[1, 1];
            ret[2, 1] = src[1, 2];
            ret[3, 1] = 0f;
            ret[0, 2] = src[2, 0];
            ret[1, 2] = src[2, 1];
            ret[2, 2] = src[2, 2];
            ret[3, 2] = 0f;
            ret[0, 3] = (-src[0, 3] * ret[0, 0]) + (-src[1, 3] * ret[0, 1]) + (-src[2, 3] * ret[0, 2]);
            ret[1, 3] = (-src[0, 3] * ret[1, 0]) + (-src[1, 3] * ret[1, 1]) + (-src[2, 3] * ret[1, 2]);
            ret[2, 3] = (-src[0, 3] * ret[2, 0]) + (-src[1, 3] * ret[2, 1]) + (-src[2, 3] * ret[2, 2]);
            ret[3, 3] = 1.0f;
            return ret;
        }
        public void Transpose() { this = Matrix.TransposedMatrix(this); }

        /*public static Matrix Invert(Matrix a)
        {
            Matrix ret;
            ret.m = new float[4, 4];
        }*/

        public static Matrix operator *(Matrix a, Matrix b)
        {
            Matrix ret;
            ret.m = new float[4, 4];
            ret[0, 0] = a[0, 0] * b[0, 0] + a[0, 1] * b[1, 0] + a[0, 2] * b[2, 0] + a[0, 3] * b[3, 0];
            ret[0, 1] = a[0, 0] * b[0, 1] + a[0, 1] * b[1, 1] + a[0, 2] * b[2, 1] + a[0, 3] * b[3, 1];
            ret[0, 2] = a[0, 0] * b[0, 2] + a[0, 1] * b[1, 2] + a[0, 2] * b[2, 2] + a[0, 3] * b[3, 2];
            ret[0, 3] = a[0, 0] * b[0, 3] + a[0, 1] * b[1, 3] + a[0, 2] * b[2, 3] + a[0, 3] * b[3, 3];
            ret[1, 0] = a[1, 0] * b[0, 0] + a[1, 1] * b[1, 0] + a[1, 2] * b[2, 0] + a[1, 3] * b[3, 0];
            ret[1, 1] = a[1, 0] * b[0, 1] + a[1, 1] * b[1, 1] + a[1, 2] * b[2, 1] + a[1, 3] * b[3, 1];
            ret[1, 2] = a[1, 0] * b[0, 2] + a[1, 1] * b[1, 2] + a[1, 2] * b[2, 2] + a[1, 3] * b[3, 2];
            ret[1, 3] = a[1, 0] * b[0, 3] + a[1, 1] * b[1, 3] + a[1, 2] * b[2, 3] + a[1, 3] * b[3, 3];
            ret[2, 0] = a[2, 0] * b[0, 0] + a[2, 1] * b[1, 0] + a[2, 2] * b[2, 0] + a[2, 3] * b[3, 0];
            ret[2, 1] = a[2, 0] * b[0, 1] + a[2, 1] * b[1, 1] + a[2, 2] * b[2, 1] + a[2, 3] * b[3, 1];
            ret[2, 2] = a[2, 0] * b[0, 2] + a[2, 1] * b[1, 2] + a[2, 2] * b[2, 2] + a[2, 3] * b[3, 2];
            ret[2, 3] = a[2, 0] * b[0, 3] + a[2, 1] * b[1, 3] + a[2, 2] * b[2, 3] + a[2, 3] * b[3, 3];
            ret[3, 0] = a[3, 0] * b[0, 0] + a[3, 1] * b[1, 0] + a[3, 2] * b[2, 0] + a[3, 3] * b[3, 0];
            ret[3, 1] = a[3, 0] * b[0, 1] + a[3, 1] * b[1, 1] + a[3, 2] * b[2, 1] + a[3, 3] * b[3, 1];
            ret[3, 2] = a[3, 0] * b[0, 2] + a[3, 1] * b[1, 2] + a[3, 2] * b[2, 2] + a[3, 3] * b[3, 2];
            ret[3, 3] = a[3, 0] * b[0, 3] + a[3, 1] * b[1, 3] + a[3, 2] * b[2, 3] + a[3, 3] * b[3, 3];

            /*for (int k = 0; k < 4; k++)
            {
                for (int l = 0; l < 4; l++)
                {

                    for (int i = 0 ; i < 4; i++)
                    {
                        float f1 = a[k, i];
                        float f2 = b[i, l];
                        
                        ret.m[k, l] += f1 * f2;
                    }
                }
            }*/

            return ret;
        }

    }

}
